"""
The vivisect CLI.
"""
import re
import sys
import shlex
import pprint
import socket
import logging
import traceback
import collections
from getopt import getopt

import vtrace
import vivisect
import vivisect.vamp as viv_vamp
import vivisect.vector as viv_vector
import vivisect.reports as viv_reports
import vivisect.tools.graphutil as viv_graph

import vivisect.tools.fscope as v_t_fscope
import vivisect.tools.graphutil as v_t_graph

import visgraph.pathcore as vg_path

import vtrace.lockstep as vt_lockstep

import vdb

import envi.exc as e_exc
import envi.cli as e_cli
import envi.common as e_common
import envi.memory as e_memory
import envi.expression as e_expr
import envi.memcanvas as e_canvas
import envi.memcanvas.renderers as e_render

from vivisect.const import *

logger = logging.getLogger(__name__)


class VivCli(e_cli.EnviCli, vivisect.VivWorkspace):
    '''
    A class that builds upon the VivWorkspace to provide command line capabilities so that
    things like the Vivisect UI can provide a cleaner interface that just a direct python
    shell. It inherits the same parameters as the VivWorkspace (autosave and confdir currently)

    To add a new command, simply add a new function called `do_<cmdname>`.
    '''
    def __init__(self, **kwargs):
        e_cli.EnviCli.__init__(self, self, symobj=self)
        vivisect.VivWorkspace.__init__(self, **kwargs)
        self.canvas.addRenderer("bytes", e_render.ByteRend())
        self.canvas.addRenderer("u_int_16", e_render.ShortRend())
        self.canvas.addRenderer("u_int_32", e_render.LongRend())
        self.canvas.addRenderer("u_int_64", e_render.QuadRend())
        import vivisect.renderers as viv_rend
        self.canvas.addRenderer("viv", viv_rend.WorkspaceRenderer(self))
        self.prompt = "viv> "
        self.addScriptPathEnvVar('VIV_SCRIPT_PATH')

    def getExpressionLocals(self):
        locs = e_cli.EnviCli.getExpressionLocals(self)
        locs['vw'] = self
        locs['vprint'] = self.vprint
        locs['vivisect'] = vivisect
        return locs

    def do_report(self, line):
        """
        Fire a report module by python path name.

        Usage: report <python.path.to.report.module>
        """
        if not line:
            self.vprint("Report Modules")
            for descr, modname in viv_reports.listReportModules():
                self.vprint("Path: %32s (Name: %s)" % (modname, descr))
            return

        cols, results = viv_reports.runReportModule(self, line)
        for va, row in results.items():
            self.canvas.addVaText("%s:\n" % self.getName(va), va)

            for indx in range(len(cols)):
                valu = row[indx]
                name, typename = cols[indx]
                self.canvas.addVaText("    " + name, va)
                self.canvas.addText(": %s\n" % valu)
            self.canvas.addText("\n")

    def do_pathcount(self, line):
        '''
        Mostly for testing the graph stuff... this will likely be removed.

        (does not count paths with loops currently...)

        Usage: pathcount <func_expr>
        '''
        if not line:
            return self.do_help('pathcount')
        try:
            fva = self.parseExpression(line)
            if not self.isFunction(fva):
                self.vprint('Not a function!')
                return

        except Exception as e:
            self.vprint(str(e))
            return

        g = v_t_graph.buildFunctionGraph(self, fva)
        pathcnt = 0
        for path in v_t_graph.getCodePaths(g):
            self.vprint('Path through 0x%.8x: %s' % (fva, [hex(p[0]) for p in path]))
            pathcnt += 1
        self.vprint('Total Paths: %d' % pathcnt)

    def do_symboliks(self, line):
        '''
        Use the new symboliks subsystem. (NOTE: i386 only for a bit...)

        Usage: symboliks [ options ]

        -A  Run the emu and show the state of the machine for all found paths
            to the given address

        '''
        if not line:
            return self.do_help("symboliks")

        watchaddr = None

        argv = e_cli.splitargs(line)
        try:
            opts, argv = getopt(argv, 'A:')
        except Exception:
            return self.do_help('symboliks')

        for opt, optarg in opts:
            if opt == '-A':
                # TODO: USE THIS
                watchaddr = self.parseExpression(optarg)

        va = self.parseExpression(argv[0])
        fva = self.getFunction(va)

        import vivisect.symboliks as viv_symboliks
        import vivisect.symboliks.common as sym_common
        import vivisect.symboliks.effects as viv_sym_effects
        import vivisect.symboliks.analysis as vsym_analysis

        symctx = vsym_analysis.getSymbolikAnalysisContext(self)

        for emu, effects in symctx.getSymbolikPaths(fva):

            self.vprint('PATH %s' % ('='*60))

            for eff in effects:

                eff.reduce(emu)
                if eff.efftype in (EFFTYPE_CONSTRAIN, EFFTYPE_CALLFUNC):
                    self.vprint(str(eff))

            for addrsym, valsym in emu._sym_mem.values():
                addrsym = addrsym.reduce(emu=emu)
                valsym = valsym.reduce(emu=emu)
                if emu.isLocalMemory(addrsym):
                    continue
                self.vprint('[ %s ] = %s' % (addrsym, valsym))
            self.vprint('RETURN  %r' % emu.getFunctionReturn().reduce())

    def do_names(self, line):
        '''
        Show any names which contain the given argument.

        Usage: names <name_regex>

        FIXME unify do_sym from vdb into symbol context!
        '''
        if not line:
            return self.do_help('names')

        regex = re.compile(line, re.I)
        for va, name in self.getNames():
            if regex.search(name):
                ftup = self.getFileAndOffset(va)
                if ftup:
                    fname, fbase, off = ftup
                    self.vprint('0x%.8x: %s  (%r + 0x%x)' % (va, name, fname, off))
                else:
                    self.vprint("WARNING: 0x%x: %r  doesn't exist in a file!" % (va, name))

    def do_save(self, line):
        """
        Save the current workspace.

        Usage: save
        """
        self.vprint("Saving workspace...")
        self.saveWorkspace()
        self.vprint("...save complete!")

    def do_xrefs(self, line):
        """
        Show xrefs for a particular location.

        Usage: xrefs [options] <va_expr>
        -T Show xrefs *to* the given address
        -F Show xrefs *from* the given address (default)
        """
        parser = e_cli.VOptionParser()
        parser.add_option('-T', action='store_true', dest='xrto')
        parser.add_option('-F', action='store_true', dest='xrfrom')
        argv = shlex.split(line)
        try:
            options, argv = parser.parse_args(argv)
        except Exception as e:
            self.vprint(repr(e))
            return self.do_help('xrefs')

        if len(argv) < 1:
            self.vprint('Supply a va_expr')
            return self.do_help('xrefs')

        va = self.parseExpression(argv[0])

        fptr = []
        if options.xrto:
            fptr.append(self.getXrefsTo)
        if options.xrfrom:
            fptr.append(self.getXrefsFrom)

        for func in fptr:
            for xrfr, xrto, rtype, rflags in func(va):
                tname = ref_type_names.get(rtype, 'Unknown')
                self.vprint('\tFrom: 0x%.8x, To: 0x%.8x, Type: %s, Flags: 0x%.8x' % (xrfr, xrto, tname, rflags))

    def do_searchopcodes(self, line):
        '''
        search opcodes/function for a pattern

        searchopcodes [-f <funcva>] [options] <pattern>
        -f [fva]   - focus on one function
        -c         - search comments
        -o         - search operands
        -t         - search text
        -M <color> - mark opcodes (default = orange)
        -R         - pattern is REGEX (otherwise just text)

        '''
        parser = e_cli.VOptionParser()
        parser.add_option('-f', action='store', dest='funcva', type='int')
        parser.add_option('-c', action='store_true', dest='searchComments')
        parser.add_option('-o', action='store_true', dest='searchOperands')
        parser.add_option('-t', action='store_true', dest='searchText')
        parser.add_option('-M', action='store', dest='markColor', default='orange')
        parser.add_option('-R', action='store_true', dest='is_regex')

        argv = shlex.split(line)
        try:
            options, args = parser.parse_args(argv)
        except Exception as e:
            self.vprint(repr(e))
            return self.do_help('searchopcodes')

        pattern = ' '.join(args)
        if len(pattern) == 0:
            self.vprint('you must specify a pattern')
            return self.do_help('searchopcodes')

        # generate our interesting va list
        valist = []
        if options.funcva:
            # setup valist from function data
            try:
                fva = options.funcva
                graph = viv_graph.buildFunctionGraph(self, fva)
            except Exception as e:
                self.vprint(repr(e))
                return

            for nva, node in graph.getNodes():
                va = nva
                endva = nva + node.get('cbsize')
                while va < endva:
                    lva, lsz, ltype, ltinfo = self.getLocation(va)
                    valist.append(va)
                    va += lsz

        else:
            # the whole workspace is our oyster
            valist = [va for va, lvsz, ltype, ltinfo in self.getLocations(LOC_OP)]

        res = []
        canv = e_canvas.StringMemoryCanvas(self)

        defaultSearchAll = True
        for va in valist:
            try:
                addthis = False
                op = self.parseOpcode(va)

                # search comment
                if options.searchComments:
                    defaultSearchAll = False
                    cmt = self.getComment(va)
                    if cmt is not None:

                        if options.is_regex:
                            if len(re.findall(pattern, cmt)):
                                addthis = True

                        else:
                            if pattern in cmt:
                                addthis = True

                # search operands
                if options.searchOperands:
                    defaultSearchAll = False
                    for opidx, oper in enumerate(op.opers):
                        # we're writing to a temp canvas, so clear it before each test
                        canv.clearCanvas()
                        oper = op.opers[opidx]
                        oper.render(canv, op, opidx)
                        operepr = canv.strval

                        if options.is_regex:
                            if len(re.findall(pattern, operepr)):
                                addthis = True

                        else:
                            if pattern in operepr:
                                addthis = True

                            # if we're doing non-regex, let's test against real numbers
                            # (instead of converting to hex and back)
                            numpattrn = pattern
                            try:
                                numpattrn = int(numpattrn, 0)
                            except:
                                pass

                            if numpattrn in vars(oper).values():
                                addthis = True

                # search full text
                if options.searchText or defaultSearchAll:
                    # search through the rendering of the opcode, as well as the comment
                    canv.clearCanvas()
                    op.render(canv)
                    oprepr = canv.strval
                    cmt = self.getComment(va)
                    if cmt is not None:
                        oprepr += "  ; " + cmt

                    if options.is_regex:
                        if len(re.findall(pattern, oprepr)):
                            addthis = True

                    else:
                        if pattern in oprepr:
                            addthis = True
                # only want one listing of each va, no matter how many times it matches
                if addthis:
                    res.append(va)
            except:
                self.vprint(''.join(traceback.format_exception(*sys.exc_info())))

        if len(res) == 0:
            self.vprint('pattern not found: %s (%s)' % (pattern.encode('utf-8').hex(), repr(pattern)))
            return

        # set the color for each finding
        color = options.markColor
        colormap = {va: color for va in res}
        if self._viv_gui is not None:
            from vqt.main import vqtevent
            vqtevent('viv:colormap', colormap)

        self.vprint('matches for: %s (%s)' % (pattern.encode('utf-8').hex(), repr(pattern)))
        for va in res:
            mbase, msize, mperm, mfile = self.memobj.getMemoryMap(va)
            pname = e_memory.reprPerms(mperm)
            sname = self.reprPointer(va)

            try:
                op = self.parseOpcode(va)
                self.canvas.renderMemory(va, len(op))
                cmt = self.getComment(va)
                if cmt is not None:
                    self.canvas.addText('\t\t; %s (Perms: %s, Smartname: %s)' % (cmt, pname, sname))

                self.canvas.addText('\n')
            except e_exc.SegmentationViolation as e:
                logger.debug("segv at 0x%x", va)

        self.vprint('done (%d results).' % len(res))

    def do_imports(self, line):
        """
        Show the imports in the workspace (or potentially just one file)

        Usage: imports [fname]
        """
        self.canvas.addText("Imports:\n")
        for va, size, ltype, tinfo in self.getImports():
            # FIXME warn them...
            # (but should we though?)
            if not tinfo.startswith(line):
                continue
            vastr = self.arch.pointerString(va)
            self.canvas.addVaText(vastr, va)
            self.canvas.addText(" ")
            self.canvas.addNameText(tinfo, tinfo)
            self.canvas.addText("\n")

    def do_fscope(self, line):
        '''
        The fscope command can be used to enumerate things from the
        scope of one function and down it's calling graph.

        Usage: fscope [options] <func_addr_expr>

        -I - Show import calls from this function scope
        -S - Show strings from this function scope

        Example: fscope -I kernel32.CreateFileW
                 (Show imports called by CreateFileW and down...)

        '''
        showimp = False
        showstr = False

        argv = e_cli.splitargs(line)
        try:
            opts, args = getopt(argv, 'IS')
        except Exception:
            return self.do_help('fscope')

        if not len(args) or not len(opts):
            return self.do_help('fscope')

        for opt, optarg in opts:
            if opt == '-I':
                showimp = True
            elif opt == '-S':
                showstr = True

        for expr in args:

            va = self.parseExpression(expr)

            if showimp:
                for callva, impname in v_t_fscope.getImportCalls(self, va):
                    pstr = self.arch.pointerString(callva)
                    self.canvas.addVaText(pstr, callva)
                    # FIXME best name symbol etc?
                    self.canvas.addText(' %s\n' % impname)

            if showstr:
                for refva, strva, strbytes in v_t_fscope.getStringRefs(self, va):
                    pstr = self.arch.pointerString(refva)
                    self.canvas.addVaText(pstr, refva)
                    self.canvas.addText(' ')
                    self.canvas.addVaText(strbytes, strva)
                    self.canvas.addText('\n')

    def do_exports(self, line):
        """
        List the exports in the workspace (or in a specific file).

        Usage: exports [fname]
        """
        edict = collections.defaultdict(list)
        for va, etype, name, filename in self.getExports():
            edict[filename].append((name, va))

        if line:
            x = edict.get(line)
            if x is None:
                self.vprint("Unknown fname: %s" % line)
                return
            edict = {line: x}

        fnames = list(edict.keys())
        fnames.sort()
        for fname in fnames:
            self.canvas.addNameText(fname, fname)
            self.canvas.addText(":\n")
            exports = edict.get(fname)
            exports.sort()
            for ename, eva in exports:
                pstr = self.arch.pointerString(eva)
                self.canvas.addText("    ")
                self.canvas.addVaText(pstr, eva)
                self.canvas.addText("  ")
                self.canvas.addNameText(ename, ename)
                self.canvas.addText("\n")

    def do_filemeta(self, line):

        '''
        Show/List file metadata.

        Usage: filemeta [ fname [ keyname ] ]

        Example: filemeta kernel32
        Example: filemeta kernel32 md5
        '''

        argv = e_cli.splitargs(line)
        if len(argv) == 0:

            self.vprint('Loaded Files:')
            for fname in self.getFiles():
                self.vprint('    %s' % fname)

        elif len(argv) == 1:
            d = self.getFileMetaDict(argv[0])
            self.vprint(pprint.pformat(d))

        elif len(argv) == 2:
            val = self.getFileMeta(argv[0], argv[1])
            self.vprint('%s (%s):' % (argv[1], argv[0]))
            self.vprint(pprint.pformat(val))

        else:
            self.do_help('filemeta')

    def do_funcmeta(self, line):
        """
        Show/Set function metadata.
        Usage: funcmeta <func_expr> [key <value_expr>]

        """
        # FIXME make a search thing here!
        argv = e_cli.splitargs(line)
        if len(argv) == 0:
            return self.do_help("funcmeta")

        if len(argv) == 1:
            va = self.parseExpression(argv[0])
            meta = self.getFunctionMetaDict(va)
            self.vprint(pprint.pformat(meta))

        elif len(argv) == 3:
            va = self.parseExpression(argv[0])
            name = argv[1]
            locs = self.getExpressionLocals()
            val = e_expr.evaluate(argv[2], locs)
            self.setFunctionMeta(va, name, val)

    def do_loc(self, line):
        """
        Display the repr of a single location by va.

        Usage: loc <va_expr>
        """
        if not line:
            return self.do_help("loc")

        addr = self.parseExpression(line)
        loc = self.getLocation(addr)
        if loc is None:
            s = self.arch.pointerString(addr)
            self.vprint("Unknown location: %s" % s)
        r = self.reprLocation(loc)
        self.vprint(r)

    def do_make(self, line):
        """
        Create new instances of locations in the vivisect workspace.

        Usage: make [options] <va_expr>
        -c Make code
        -f Make function
        -s Make a string
        -u Make a unicode string
        -n <size> Make a number
        -p <size> Make a pad
        -S <structname> Make a structure
        """
        argv = e_cli.splitargs(line)
        try:
            opts, args = getopt(argv, "csufn:p:S:")
        except Exception as e:
            logger.warning(str(e))
            return self.do_help("make")

        if len(args) != 1 or len(opts) != 1:
            return self.do_help("make")

        addr = self.parseExpression(args[0])
        opt, optarg = opts[0]

        if opt == "-f":
            logger.debug('new function (manual-cli): 0x%x', addr)
            self.makeFunction(addr)

        elif opt == "-c":
            self.makeCode(addr)

        elif opt == "-s":
            self.makeString(addr)

        elif opt == "-u":
            self.makeUnicode(addr)

        elif opt == "-n":
            size = self.parseExpression(optarg)
            self.makeNumber(addr, size)

        elif opt == "-p":
            size = self.parseExpression(optarg)
            self.makePad(addr, size)

        elif opt == "-S":
            self.makeStructure(addr, optarg)

        else:
            return self.do_help("make")

    def do_emulate(self, line):
        """
        Create an emulator for the given function, and drop into a vdb
        interface to step through the code.

        (vdb CLI will appear in controlling terminal...)

        Usage: emulate <va_expr>
        """
        if not line:
            return self.do_help("emulate")

        addr = self.parseExpression(line)
        emu = self.getEmulator(va=addr)

        trace = vt_lockstep.TraceEmulator(emu)

        db = vdb.Vdb(trace=trace)
        db.cmdloop()

    def _do_argtrack(self, line):
        #FIXME: this is currently broken and needs to be revamped to use 
        # apiCall and VivTaints
        """
        Track input arguments to the given function by name or address.

        (this is currently broken.  sorry!)

        Usage: argtrack <func_addr_expr> <arg_idx>
        """
        if not line:
            return self.do_help("argtrack")

        argv = e_cli.splitargs(line)
        if len(argv) != 2:
            return self.do_help("argtrack")

        try:
            fva = self.parseExpression(argv[0])
        except Exception:
            self.vprint("Invalid Address Expression: %s" % argv[0])
            return

        try:
            idx = self.parseExpression(argv[1])
        except Exception:
            self.vprint("Invalid Index Expression: %s" % argv[1])
            return

        if self.getFunction(fva) != fva:
            self.vprint("Invalid Function Address: (0x%.8x) %s" % (fva, line))

        for pleaf in viv_vector.trackArgOrigin(self, fva, idx):

            self.vprint('='*80)

            path = vg_path.getPathToNode(pleaf)
            path.reverse()

            for pnode in path:
                fva = vg_path.getNodeProp(pnode, 'fva')
                argv = vg_path.getNodeProp(pnode, 'argv')
                callva = vg_path.getNodeProp(pnode, 'cva')
                argidx = vg_path.getNodeProp(pnode, 'argidx')
                if callva is not None:
                    aval, amagic = argv[argidx]
                    arepr = '0x%.8x' % aval
                    if amagic is not None:
                        arepr = repr(amagic)
                    frepr = 'UNKNOWN'
                    if fva is not None:
                        frepr = '0x%.8x' % fva
                    self.vprint('func: %s calls at: 0x%.8x with his own: %s' % (frepr, callva, arepr))
            self.vprint("="*80)

    def do_chat(self, line):
        """
        Echo a message to any other users of a shared workspace.

        Usage: chat oh hai! Checkout 0x7c778030
        """
        if len(line) == 0:
            return self.do_help('chat')

        self.chat(line)

    def do_codepath(self, line):
        """
        Enumerate and show any known code paths from the specified
        from address expression to the to address expression.
        Usage: codepath <from_expr> <to_expr>
        """
        if not line:
            return self.do_help("codepath")

        argv = e_cli.splitargs(line)
        if len(argv) != 2:
            return self.do_help("codepath")

        try:
            frva = self.parseExpression(argv[0])
        except Exception:
            self.vprint("Invalid From Va: %s" % argv[0])
            return

        try:
            tova = self.parseExpression(argv[1])
        except Exception:
            self.vprint("Invalid To Va: %s" % argv[1])
            return

        self.vprint("Tracking Paths From 0x%.8x to 0x%.8x" % (frva, tova))

        paths = viv_vector.getCodePaths(self, frva, tova)
        self.vprint("Function VA\tBlock VA\tSize\tFunction Name")
        count = 0
        for blist in paths:
            count += 1
            self.vprint("="*30)
            for bva, bsize, fva in blist:
                fname = self.getName(fva)
                self.vprint("0x%.8x\t0x%.8x\t%4d\t%s" % (fva, bva, bsize, fname))
        if count == 0:
            self.vprint("None!")
            return

    def do_vampsig(self, line):
        """
        Generate a vamp signature string for the given function's first block.
        """
        if not line:
            return self.do_help("vampsig")

        va = self.parseExpression(line)

        fva = self.getFunction(va)
        if fva is None:
            self.vprint("Invalid Function Address: 0x%.8x (%s)" % (va, line))

        sig, mask = viv_vamp.genSigAndMask(self, fva)
        self.vprint("SIGNATURE: %s" % e_common.hexify(sig))
        self.vprint("MASK: %s" % e_common.hexify(mask))

    def do_vdb(self, line):
        '''
        Execute vdb GUI from within vivisect (allowing special hooks between them...)
        (Optionally, specify a host to use for remote vdb debugging)

        Usage: vdb [<remote_host>]
        '''
        if line:
            try:
                socket.gethostbyname(line)
            except Exception:
                self.vprint('Invalid Remote Host: %s' % line)

            vtrace.remote = line

        import vivisect.vdbext as viv_vdbext
        viv_vdbext.runVdb(self._viv_gui)

    def do_switch(self, line):
        '''
        Wire up a switch case.  

        Usage: switch <jmp_va> <count> [offset]
            where:
                jmp_va  - the va of the  "jmp reg" instruction
                count   - number of switch case indices are covered by this switch case jmp
                offset  - first switch case index.
            
            eg: switch case handles i == 32 thru 47, count = 16, offset = 32
        '''
        if not line:
            return self.do_help("switch")

        argv = e_cli.splitargs(line)
        if len(argv) < 2:
            return self.do_help("switch")

        offset = 0
        try:
            if argv[0].startswith("0x"):
                jmpva = int(argv[0], 16)
            else:
                jmpva = int(argv[0])

            if argv[1].startswith("0x"):
                count = int(argv[1], 16)
            else:
                count = int(argv[1])

            if len(argv) == 3:
                if argv[2].startswith("0x"):
                    offset = int(argv[2], 16)
                else:
                    offset = int(argv[2])

            import vivisect.analysis.generic.symswitchcase as vagss
            import vivisect.analysis.generic.codeblocks as vagc
            sc = vagss.SwitchCase(self, jmpva)
            sc.makeSwitch(count, offset)
            funcva = self.getFunction(jmpva)
            vagc.analyzeFunction(self, funcva)
        except ValueError as e:
            self.vprint("exception: %r" % e)
            return self.do_help("switch")

    def do_m_switch(self, line):
        '''
        Wire up a switch-case based on an array of pointers

        Usage: m_switch <jmp_va> <array_va> <count> [baseva] [offset] [size]
            where:
                jmp_va  - the va of the  "jmp reg" instruction
                array_va - the location of the first pointer in the array
                count   - number of switch case indices are covered by this switch case jmp
                baseva  - add the offset to this address (eg. _GOT_)
                offset  - first switch case index.
                [size]  - size (default: pointer size)
            
            eg: switch case handles i == 32 thru 47, count = 16, offset = 32
        '''
        if not line:
            return self.do_help("m_switch")

        argv = e_cli.splitargs(line)
        if len(argv) < 3:
            return self.do_help("m_switch")

        size = 0
        baseva = 0
        offset = 0
        try:
            jmpva = self.parseExpression(argv[0])
            array = self.parseExpression(argv[1])
            count = self.parseExpression(argv[2])

            if len(argv) > 3:
                baseva = self.parseExpression(argv[3])

            if len(argv) > 4:
                offset = self.parseExpression(argv[4])

            if len(argv) > 5:
                size = self.parseExpression(argv[5])

            # link up the switchcase using data provided
            import vivisect.analysis.generic.symswitchcase as vagss
            import vivisect.analysis.generic.codeblocks as vagc
            vagss.link_up(self, jmpva, array, count, offset, baseva, size)

            # reanalyze codeblocks for the function (since it should have changed)
            funcva = self.getFunction(jmpva)
            if funcva is not None:
                vagc.analyzeFunction(self, funcva)

        except ValueError as e:
            self.vprint("exception: %r" % e)
            return self.do_help("m_switch")

    def do_plt(self, line):
        '''
        Parse an entire PLT Section

        Usage: plt <pltva> <pltsize>
        '''
        if not line:
            return self.do_help("plt")

        argv = e_cli.splitargs(line)
        if len(argv) != 2:
            return self.do_help("plt")

        sva = self.parseExpression(argv[0])
        ssize = self.parseExpression(argv[1])

        import vivisect.analysis.elf.elfplt as vaee
        vaee.analyzePLT(self, sva, ssize)

    def do_plt_function(self, line):
        '''
        Make a PLT function at a virtual address

        Usage: plt_function <va>
        '''
        if not line:
            return self.do_help("plt_function")

        fva = self.parseExpression(line)

        import vivisect.analysis.elf.elfplt as vaee
        vaee.analyzeFunction(self, fva)

    def do_leaders(self, line):
        '''
        Manage Leader Sessions.  This is useful if sessions are left hanging 
        if a leader disconnects.
        
        Usage: leaders (list,kill,killall)

            list - show leader sessions
            kill <uuidhex> - kill a particular session
            killall - kill all leader sessions for a given binary
            mod <uuid> <user> <sessionname> - rename a leader session/user

        Example:
            leaders kill 197a7797ba1511ecbe41091ba860c051
        '''
        if not line:
            return self.do_help("leaders")

        argv = e_cli.splitargs(line)
        try:
            if argv[0] == 'list':
                self.vprint("Current Leader Sessions:")
                for uuid, (user, fname) in self.getLeaderSessions().items():
                    curva = self.getLeaderLoc(uuid)
                    self.vprint(" * %s (user: %s  uuid: %s)" % (fname, user, uuid))

            elif argv[0] == 'mod':
                uuid = argv[1]
                user = argv[2]
                if len(argv) > 4:
                    sessionname = ' '.join(argv[3:])
                else:
                    sessionname = argv[3]

                self.modifyLeaderSession(uuid, user, sessionname)

            elif argv[0] == 'kill':
                if len(argv) < 2:
                    self.vprint('kill requires a valid <uuid> argument')

                uuid = argv[1]
                self.killLeaderSession(uuid)
                
            elif argv[0] == 'killall':
                for uuid, (user, fname) in self.getLeaderSessions().items():
                    self.vprint("*killing %r (%r, %r)" % (uuid, user, fname))
                    self.killLeaderSession(uuid)

        except Exception as e:
            self.vprint("Error: %r" % e)



'''
A place for some no-brainer basics :)
'''
from PyQt5 import QtCore
from PyQt5.QtWidgets import *


class BasicTreeView(QTreeView):

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent=parent)
        self.setAlternatingRowColors( True )
        self.setSortingEnabled( True )

    def setModel(self, model):
        ret = QTreeView.setModel(self, model)
        c = len(model.columns)
        for i in range(c):
            self.resizeColumnToContents(i)
        return ret

class BasicModel(QtCore.QAbstractItemModel):

    columns = ('one','two')

    def __init__(self, rows=()):
        QtCore.QAbstractItemModel.__init__(self)
        self.rows = list(rows)

    def index(self, row, column, parent):
        return self.createIndex(row, column, self.rows[row])

    def parent(self, index):
        return QtCore.QModelIndex()

    def rowCount(self, index):
        if index.internalPointer() in self.rows:
            return 0
        return len(self.rows)

    def data(self, index, role):
        if role == 0:
            row = index.row()
            col = index.column()
            return self.rows[row][col]

        else:
            return None

    def sort(self, col, order=QtCore.Qt.AscendingOrder):
        self.layoutAboutToBeChanged.emit()
        self.rows.sort( cmp=lambda x,y: cmp(x[col],y[col]) )
        if order == QtCore.Qt.DescendingOrder:
            self.rows.reverse()
        self.layoutChanged.emit()

    def columnCount(self, index):
        return len(self.columns)

    def headerData(self, column, orientation, role):

        if ( orientation == QtCore.Qt.Horizontal and
             role == QtCore.Qt.DisplayRole):

            return self.columns[column]

        return None

class VBox( QVBoxLayout ):

    def __init__(self, *widgets):
        QVBoxLayout.__init__(self)
        self.setContentsMargins(2,2,2,2)
        self.setSpacing(4)
        for w in widgets:
            if w is None:
                self.addStretch()
                continue
            self.addWidget( w )

class HBox( QHBoxLayout ):

    def __init__(self, *widgets):
        QHBoxLayout.__init__(self)
        self.setContentsMargins(2,2,2,2)
        self.setSpacing(4)
        for w in widgets:
            if w is None:
                self.addStretch()
                continue
            self.addWidget( w )

class ACT:
    def __init__(self, meth, *args, **kwargs):
        self.meth = meth
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.meth( *self.args, **self.kwargs )




import sys
import argparse

import PE
import vtrace.platforms.win32 as vt_win32
import vstruct.builder as vs_builder


def setup():
    ap = argparse.ArgumentParser(help='Dump PE structures')
    ap.add_argument('file')
    return ap


def main(argv):
    opts = setup().parse_args(argv)
    with open(opts.file, 'rb') as f:
        p = PE.PE(f)
        baseaddr = p.IMAGE_NT_HEADERS.OptionalHeader.ImageBase
        osmajor = p.IMAGE_NT_HEADERS.OptionalHeader.MajorOperatingSystemVersion
        osminor = p.IMAGE_NT_HEADERS.OptionalHeader.MinorOperatingSystemVersion
        machine = p.IMAGE_NT_HEADERS.FileHeader.Machine

        vsver = p.getVS_VERSIONINFO()

        archname = PE.machine_names.get(machine)

        parser = vt_win32.Win32SymbolParser(0xffffffff, opts.file, baseaddr)
        parser.parse()

        t = parser._sym_types.values()
        e = parser._sym_enums.values()
        builder = vs_builder.VStructBuilder(defs=t, enums=e)

        print('# Version: %d.%d' % (osmajor, osminor))
        print('# Architecture: %s' % archname)
        if vsver is not None:
            keys = vsver.getVersionKeys()
            keys.sort()
            for k in keys:
                val = vsver.getVersionValue(k)
                print('# %s: %s' % (k, val))
        print(builder.genVStructPyCode())


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import string
import binascii

import envi.const as e_const
import envi.common as e_common
import envi.memory as e_memory
import envi.memcanvas as e_canvas
import envi.memcanvas.renderers as e_render

from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import *


class VQLineEdit(QLineEdit):
    '''
    Has an additional signal to emit a signal on release of every keypress.
    '''
    keyReleased = QtCore.pyqtSignal(QtGui.QKeyEvent)

    def keyReleaseEvent(self, event):
        self.keyReleased.emit(event)
        QLineEdit.keyReleaseEvent(self, event)


class MemNavWidget(QWidget):
    userChanged = QtCore.pyqtSignal(str, str)

    def __init__(self):
        QWidget.__init__(self)

        self.expr_entry = QLineEdit()
        self.esize_entry = QLineEdit()

        hbox1 = QHBoxLayout()
        hbox1.setContentsMargins(2, 2, 2, 2)
        hbox1.setSpacing(4)
        hbox1.addWidget(self.expr_entry)
        hbox1.addWidget(self.esize_entry)

        self.setLayout(hbox1)

        self.expr_entry.returnPressed.connect(self.emitUserChangedSignal)
        self.esize_entry.returnPressed.connect(self.emitUserChangedSignal)

    def emitUserChangedSignal(self):
        '''
        Emits signal when user manually enters new expressions in the expr or
        size field and presses enter.
        '''
        expr = str(self.expr_entry.text())
        size = str(self.esize_entry.text())
        self.userChanged.emit(expr, size)

    def setValues(self, expr, esize):
        '''
        Called externally to allow programmatic way to update the expr or size
        field. Does not emit the changed signal.
        '''
        self.expr_entry.setText(expr)
        self.esize_entry.setText(esize)

    def getValues(self):
        return str(self.expr_entry.text()), str(self.esize_entry.text())

class MemWriteWindow(QWidget):
    '''
    gui for writemem cli command.
    '''
    renderRequest = QtCore.pyqtSignal(str, str)

    # button to write memory was clicked (va, bytez)
    writeToMemory = QtCore.pyqtSignal(str, str)

    def __init__(self, expr='', esize='', emu=None, parent=None):
        QWidget.__init__(self, parent=parent)

        self.modes = ['ascii', 'hex', 'regex', 'utf-8', 'utf-16-le',
                        'utf-16-be']

        rend_orig = e_render.ByteRend()
        self.canvas_orig = e_canvas.StringMemoryCanvas(None)
        self.canvas_orig.addRenderer('bytes', rend_orig)

        rend_new = e_render.ByteRend()
        self.canvas_new = e_canvas.StringMemoryCanvas(None)
        self.canvas_new.addRenderer('bytes', rend_new)

        hbox1 = QHBoxLayout()
        self.nav = MemNavWidget()
        self.nav.userChanged.connect(self.renderMemory)
        self.renderRequest.connect(self.nav.setValues)
        hbox1.addWidget(self.nav)

        hbox2 = QHBoxLayout()
        self.hex_edit = QPlainTextEdit()
        self.hex_edit.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.hex_edit.setReadOnly(True)
        font = QtGui.QFont('Courier') # should use actual memcanvas
        self.hex_edit.setFont(font)
        hbox2.addWidget(self.hex_edit)

        vbox1 = QVBoxLayout()
        vbox1.addLayout(hbox1)
        vbox1.addLayout(hbox2)
        gbox1 = QGroupBox('Original Bytes')
        gbox1.setLayout(vbox1)

        hbox3 = QHBoxLayout()
        mode_label = QLabel('Input:')
        self.mode_combo = QComboBox()
        self.mode_combo.addItems(self.modes)
        self.mode_combo.currentIndexChanged.connect(self.encodingChanged)
        hbox3.addWidget(mode_label)
        hbox3.addWidget(self.mode_combo, alignment=QtCore.Qt.AlignLeft)
        hbox3.addStretch(1)

        hbox4 = QHBoxLayout()
        data_label = QLabel('Bytes:')
        self.data_edit = VQLineEdit()
        self.data_edit.keyReleased.connect(self.keyReleasedSlot)
        hbox4.addWidget(data_label)
        hbox4.addWidget(self.data_edit)

        vbox2 = QVBoxLayout()
        vbox2.addLayout(hbox3)
        vbox2.addLayout(hbox4)
        gbox2 = QGroupBox('New Bytes')
        gbox2.setLayout(vbox2)

        hbox5 = QHBoxLayout()
        self.hex_preview = QPlainTextEdit()
        self.hex_preview.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.hex_preview.setReadOnly(True)
        self.hex_preview.setFont(font)
        hbox5.addWidget(self.hex_preview)

        vbox3 = QVBoxLayout()
        vbox3.addLayout(hbox5)
        gbox3 = QGroupBox('Result Preview')
        gbox3.setLayout(vbox3)

        hbox6 = QHBoxLayout()
        button = QPushButton('Write Memory')
        button.clicked.connect(self.buttonClicked)
        hbox6.addWidget(button)

        vbox = QVBoxLayout()
        vbox.addWidget(gbox1)
        vbox.addWidget(gbox2)
        vbox.addWidget(gbox3)
        vbox.addLayout(hbox6)
        self.setLayout(vbox)

        self.setWindowTitle('Memory Write')
        self.resize(650, 500)
        self.data_edit.setFocus()
        self.emu = emu
        self.renderMemory(expr, esize)

    def renderMemory(self, expr=None, esize=None, emu=None):
        if emu is not None:
            self.emu = emu

        curexpr, cur_esize = self.nav.getValues()
        if expr is None:
            expr = curexpr
        if esize is None:
            esize = cur_esize

        self.renderRequest.emit(expr, esize)

        try:
            # str() for QString -> ascii strings
            va = self.emu.parseExpression(str(expr))
            size = self.emu.parseExpression(str(esize))
            bytez = self.emu.readMemory(va, size)

            self.updateHexOrig(va, bytez)

            encoding = str(self.mode_combo.currentText())
            rbytes = str(self.data_edit.text())
            erbytes = self.encodeData(rbytes, encoding)

            # encoded bytes is bigger than the amount we are displaying.
            if len(erbytes) > len(bytez):
                self.hex_preview.setPlainText('too many bytes, change size, encoding or input')
                return

            bytez = erbytes + bytez[len(erbytes):]
            self.updateHexPreview(va, bytez)
        except Exception as e:
            self.hex_preview.setPlainText(str(e))

    def keyReleasedSlot(self, event):
        encoding = self.mode_combo.currentText()
        self.encodingChanged(None)

    def encodingChanged(self, idx):
        encoding = str(self.mode_combo.currentText())

        validator = None
        if encoding == 'hex':
            # only clear the box if there are non-hex chars
            # before setting the validator.
            txt = str(self.data_edit.text())
            if not all(c in string.hexdigits for c in txt):
                self.data_edit.setText('')

            regex = QtCore.QRegExp('^[0-9A-Fa-f]+$')
            validator = QtGui.QRegExpValidator(regex)

        self.data_edit.setValidator(validator)

        self.renderMemory()

    def encodeData(self, txt, encoding):
        if encoding == 'hex' and (len(txt) % 2) != 0:
            txt = txt[:-1] # trim last if odd length

        if encoding == 'hex':
            if not all(c in string.hexdigits for c in txt):
                return None

            return txt.decode(encoding)

        elif encoding == 'regex':
            return None

        return txt.encode(encoding)

    def updateHexOrig(self, va, bytez):
        if bytez is None:
            self.hex_edit.setPlainText('')
            return

        self.canvas_orig.clearCanvas()
        mem = e_memory.MemoryObject()
        mem.addMemoryMap(va, e_const.MM_READ, b'', bytez)
        self.canvas_orig.mem = mem
        self.canvas_orig.renderMemory(va, len(bytez))
        self.hex_edit.setPlainText(str(self.canvas_orig))

    def updateHexPreview(self, va, bytez):
        if bytez is None:
            self.hex_preview.setPlainText('')
            return

        self.canvas_new.clearCanvas()
        mem = e_memory.MemoryObject()
        mem.addMemoryMap(va, e_const.MM_READ, b'', bytez)
        self.canvas_new.mem = mem
        self.canvas_new.renderMemory(va, len(bytez))
        self.hex_preview.setPlainText(str(self.canvas_new))

    def buttonClicked(self):
        curexpr, cur_esize = self.nav.getValues()
        encoding = str(self.mode_combo.currentText())
        rbytes = str(self.data_edit.text())
        erbytes = self.encodeData(rbytes, encoding)

        hexbytes = e_common.hexify(erbytes)
        self.writeToMemory.emit(curexpr, hexbytes)

    def getValues(self):
        return self.nav.getValues()

    def setValues(self, expr, esize):
        self.nav.setValues(expr, esize)

class MockEmu(object):
    def parseExpression(self, expr):
        return int(eval(expr, {}, {}))

    def readMemory(self, va, size):
        return '\x90' * size

def main():
    import sys

    app = QApplication([])
    w = MemWriteWindow('0x1234', '0xff', emu=MockEmu())
    w.show()

    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



import logging
from io import BytesIO

import PE
import PE.carve as pe_carve

import vstruct
import vivisect
import vivisect.exc as v_exc
import vivisect.parsers as v_parsers
# Steal symbol parsing from vtrace
import vtrace  # needed only for setting the logging level
import vtrace.platforms.win32 as vt_win32

import envi.exc as e_exc
import envi.bits as e_bits
import envi.const as e_const
import envi.symstore.symcache as e_symcache

from vivisect.const import *

logger = logging.getLogger(__name__)

for mod in (PE, vtrace):
    olog = logging.getLogger(mod.__name__)
    olog.setLevel(logger.getEffectiveLevel())

# PE Machine field values
# 0x14d   Intel i860
# 0x14c   Intel I386 (same ID used for 486 and 586)
# 0x162   MIPS R3000
# 0x166   MIPS R4000
# 0x183   DEC Alpha AXP


def parseFile(vw, filename, baseaddr=None):
    pe = PE.PE(open(filename, "rb"))
    return loadPeIntoWorkspace(vw, pe, filename=filename, baseaddr=baseaddr)


def parseBytes(vw, bytes, baseaddr=None):
    fd = BytesIO(bytes)
    fd.seek(0)
    pe = PE.PE(fd)
    return loadPeIntoWorkspace(vw, pe, baseaddr=baseaddr)


def parseMemory(vw, memobj, base):
    pe = PE.peFromMemoryObject(memobj, base)
    # FIXME does the PE's load address get fixedup on rebase?
    return loadPeIntoWorkspace(vw, pe, filename=None)


def parseFd(vw, fd, filename=None, baseaddr=None):
    fd.seek(0)
    pe = PE.PE(fd)
    return loadPeIntoWorkspace(vw, pe, filename=filename, baseaddr=baseaddr)


arch_names = {
    PE.IMAGE_FILE_MACHINE_I386: 'i386',
    PE.IMAGE_FILE_MACHINE_AMD64: 'amd64',
    PE.IMAGE_FILE_MACHINE_ARM: 'arm',
    # DEV: uncomment this line to enable the arch (prolly not going to happen until AArch64 gets finished)
    # PE.IMAGE_FILE_MACHINE_ARM64: 'arm64',
    PE.IMAGE_FILE_MACHINE_ARMNT: 'thumb',
    PE.IMAGE_FILE_MACHINE_THUMB: 'thumb16',
}

archcalls = {
    'i386': 'cdecl',
    'amd64': 'msx64call',
    'arm': 'armcall',
    'thumb': 'armcall',
    'thumb16': 'armcall',
}

# map PE relocation types to vivisect types where possible
relmap = {
    PE.IMAGE_REL_BASED_HIGHLOW: (vivisect.RTYPE_BASEOFF, 4),
    PE.IMAGE_REL_BASED_DIR64: (vivisect.RTYPE_BASEOFF, 8),
}


def loadPeIntoWorkspace(vw, pe, filename=None, baseaddr=None):
    # get baseaddr and size, then make sure we have a good baseaddr
    filebaseaddr, size = getMemBaseAndSize(vw, pe, baseaddr=baseaddr)
    if baseaddr is None:
        baseaddr = filebaseaddr

    logger.debug('initial file baseva: 0x%x  size: 0x%x', baseaddr, size)
    baseaddr = vw.findFreeMemoryBlock(size, baseaddr)
    logger.info("loadPeIntoWorkspace:  loading %r (size: 0x%x) at 0x%x", filename, size, baseaddr)

    mach = pe.IMAGE_NT_HEADERS.FileHeader.Machine

    arch = arch_names.get(mach)
    if arch is None:
        raise v_exc.InvalidArchitecture("PE", hex(mach))

    vw.setMeta('Architecture', arch)
    vw.setMeta('Format', 'pe')
    vw.parsedbin = pe
    ### TODO: this doesn't work from Memory (read() needs to know how big the file is)
    byts = pe.getFileBytes()
    vw.setMeta('FileBytes', v_parsers.compressBytes(byts))

    platform = 'windows'

    # Drivers are platform "winkern" so impapi etc works
    subsys = pe.IMAGE_NT_HEADERS.OptionalHeader.Subsystem
    if subsys == PE.IMAGE_SUBSYSTEM_NATIVE:
        platform = 'winkern'

    vw.setMeta('Platform', platform)

    vw.setMeta('DefaultCall', archcalls.get(arch, 'unknown'))
    logger.info("PE loader: Arch: %r\tFormat: pe\tPlatform: %r\tFilename: %r\tBaseAddr: 0x%x", \
            arch, platform, filename, baseaddr)

    # Set ourselves up for extended windows binary analysis

    imagebase = pe.IMAGE_NT_HEADERS.OptionalHeader.ImageBase
    if baseaddr is None:
        baseaddr = imagebase

    entryrva = pe.IMAGE_NT_HEADERS.OptionalHeader.AddressOfEntryPoint
    entry = entryrva + baseaddr

    codebase = pe.IMAGE_NT_HEADERS.OptionalHeader.BaseOfCode
    codesize = pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfCode
    codervamax = codebase+codesize

    logger.info("PE Imagebase: 0x%x\tentry: 0x%x\tcodebase: 0x%x\tcodesize: 0x%x", \
            imagebase, entry, codebase, codesize)
    # grab the file bytes for hashing
    pe.fd.seek(0)
    fhash = v_parsers.md5Bytes(byts)
    sha256 = v_parsers.sha256Bytes(byts)

    fvivname = filename
    # This will help linkers with files that are re-named
    dllname = pe.getDllName()
    if dllname:
        fvivname = dllname

    if fvivname is None:
        fvivname = fhash

    logger.info("PE dllname: %r\tfvivname: %r\tmd5: %r\tsha256: %r", \
            dllname, fvivname, fhash, sha256)

    # create the file and store md5 and sha256 hashes
    fname = vw.addFile(fvivname.lower(), baseaddr, fhash)
    vw.setFileMeta(fname, 'sha256', sha256)

    symhash = e_symcache.symCacheHashFromPe(pe)
    vw.setFileMeta(fname, 'SymbolCacheHash', symhash)

    # Add file version info if VS_VERSIONINFO has it
    try:
        vs = pe.getVS_VERSIONINFO()
    except Exception as e:
        vs = None
        vw.vprint('Failed to load version info resource due to %s' % (repr(e),))
    if vs is not None:
        vsver = vs.getVersionValue('FileVersion')
        if vsver is not None and len(vsver):
            # add check to split seeing samples with spaces and nothing else..
            parts = vsver.split()
            if len(parts):
                vsver = vsver.split()[0]
                vw.setFileMeta(fname, 'Version', vsver)

    # Setup some va sets used by windows analysis modules
    # these will already exist for Multi-file workspaces
    vaSetNames = vw.getVaSetNames()
    if not 'Library Loads' in vaSetNames:
        vw.addVaSet("Library Loads", (("Address", VASET_ADDRESS), ("Library", VASET_STRING)))
    if not 'pe:ordinals' in vaSetNames:
        vw.addVaSet('pe:ordinals', (('Address', VASET_ADDRESS), ('Ordinal', VASET_INTEGER)))
    if not 'DelayImports' in vaSetNames:
        vw.addVaSet('DelayImports', (('Address', VASET_ADDRESS), ('DelayImport', VASET_STRING)))

    # SizeOfHeaders spoofable...
    curr_offset = pe.IMAGE_DOS_HEADER.e_lfanew + len(pe.IMAGE_NT_HEADERS)

    secsize = len(vstruct.getStructure("pe.IMAGE_SECTION_HEADER"))

    sec_offset = pe.IMAGE_DOS_HEADER.e_lfanew + 4 + len(pe.IMAGE_NT_HEADERS.FileHeader) + pe.IMAGE_NT_HEADERS.FileHeader.SizeOfOptionalHeader 

    if sec_offset != curr_offset:
        header_size = sec_offset + pe.IMAGE_NT_HEADERS.FileHeader.NumberOfSections * secsize
    else:
        header_size = pe.IMAGE_DOS_HEADER.e_lfanew + len(pe.IMAGE_NT_HEADERS) + pe.IMAGE_NT_HEADERS.FileHeader.NumberOfSections * secsize

    # Add the first page mapped in from the PE header.
    header = pe.readAtOffset(0, header_size)

    if not header:
        raise v_exc.CorruptPeFile("truncated PE header")

    secalign = pe.IMAGE_NT_HEADERS.OptionalHeader.SectionAlignment
    filealign = pe.IMAGE_NT_HEADERS.OptionalHeader.FileAlignment
    subsys_majver = pe.IMAGE_NT_HEADERS.OptionalHeader.MajorSubsystemVersion
    subsys_minver = pe.IMAGE_NT_HEADERS.OptionalHeader.MinorSubsystemVersion

    if secalign == 0:
        raise v_exc.CorruptPeFile("section alignment is zero")

    secrem = len(header) % secalign
    if secrem != 0:
        header += b'\x00' * (secalign - secrem)

    vw.addMemoryMap(baseaddr, e_const.MM_READ, fname, header)
    vw.addSegment(baseaddr, len(header), "PE_Header", fname)

    hstruct = vw.makeStructure(baseaddr, "pe.IMAGE_DOS_HEADER")
    magicaddr = hstruct.e_lfanew
    if vw.readMemory(baseaddr + magicaddr, 2) != b"PE":
        raise Exception("We only support PE exe's")

    if not vw.isLocation(baseaddr + magicaddr):
        padloc = vw.makePad(baseaddr + magicaddr, 4)

    ifhdr_va = baseaddr + magicaddr + 4
    ifstruct = vw.makeStructure(ifhdr_va, "pe.IMAGE_FILE_HEADER")
    ohstruct = vw.makeStructure(ifhdr_va + len(ifstruct), "pe.IMAGE_OPTIONAL_HEADER")
    nxcompat = ohstruct.DllCharacteristics & PE.IMAGE_DLLCHARACTERISTICS_NX_COMPAT

    # get resource data directory
    ddir = pe.getDataDirectory(PE.IMAGE_DIRECTORY_ENTRY_RESOURCE)
    loadrsrc = vw.config.viv.parsers.pe.loadresources
    carvepes = vw.config.viv.parsers.pe.carvepes

    deaddirs = [PE.IMAGE_DIRECTORY_ENTRY_EXPORT,
                PE.IMAGE_DIRECTORY_ENTRY_IMPORT,
                PE.IMAGE_DIRECTORY_ENTRY_RESOURCE,
                PE.IMAGE_DIRECTORY_ENTRY_EXCEPTION,
                PE.IMAGE_DIRECTORY_ENTRY_SECURITY,
                PE.IMAGE_DIRECTORY_ENTRY_BASERELOC,
                PE.IMAGE_DIRECTORY_ENTRY_DEBUG,
                PE.IMAGE_DIRECTORY_ENTRY_COPYRIGHT,
                PE.IMAGE_DIRECTORY_ENTRY_ARCHITECTURE,
                PE.IMAGE_DIRECTORY_ENTRY_GLOBALPTR,
                PE.IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG,
                PE.IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT,
                PE.IMAGE_DIRECTORY_ENTRY_IAT,
                PE.IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT,
                PE.IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR]
    deadvas = [ddir.VirtualAddress]
    for datadir in deaddirs:
        d = pe.getDataDirectory(datadir)
        if d.VirtualAddress:
            deadvas.append(d.VirtualAddress)

    for idx, sec in enumerate(pe.sections):
        mapflags = 0
        offset = sec.vsGetMeta("Offset", None)
        if offset:
            addr = baseaddr + offset
            vw.makeStructure(addr, 'pe.IMAGE_SECTION_HEADER')
        chars = sec.Characteristics
        if chars & PE.IMAGE_SCN_MEM_READ:
            mapflags |= e_const.MM_READ

            isrsrc = (sec.VirtualAddress == ddir.VirtualAddress)
            if isrsrc and not loadrsrc:
                continue

            # If it's for an older system, just about anything
            # is executable...
            # However, there is the DLLCHARACTERISTICS NXCOMPAT flag to take into account,
            # which works with the OS to prevent certain pages of memory from achieving 
            # execution unless they're marked with the execute bit
            # so we can't just blindly mark these as executable quite yet.
            if not nxcompat:
                if not vw.config.viv.parsers.pe.nx and subsys_majver < 6 and not isrsrc:
                    mapflags |= e_const.MM_EXEC

        if chars & PE.IMAGE_SCN_MEM_READ:
            mapflags |= e_const.MM_READ
        if chars & PE.IMAGE_SCN_MEM_WRITE:
            mapflags |= e_const.MM_WRITE
        if chars & PE.IMAGE_SCN_MEM_EXECUTE:
            mapflags |= e_const.MM_EXEC
        if chars & PE.IMAGE_SCN_CNT_CODE:
            mapflags |= e_const.MM_EXEC
        if chars & PE.IMAGE_SCN_CNT_UNINITIALIZED_DATA:
            mapflags |= e_const.MM_UNINIT

        secrva = sec.VirtualAddress
        secvsize = sec.VirtualSize
        secfsize = sec.SizeOfRawData
        secbase = secrva + baseaddr
        secname = sec.Name.strip("\x00")
        secrvamax = secrva + secvsize

        # If the section is part of BaseOfCode->SizeOfCode
        # force execute perms...
        if secrva >= codebase and secrva < codervamax:
            mapflags |= e_const.MM_EXEC

        # If the entry point is in this section, force execute
        # permissions.
        if secrva <= entryrva and entryrva < secrvamax:
            mapflags |= e_const.MM_EXEC

        if not nxcompat:
            if not vw.config.viv.parsers.pe.nx and subsys_majver < 6 and mapflags & e_const.MM_READ:
                mapflags |= e_const.MM_EXEC

        if sec.VirtualSize == 0 or sec.SizeOfRawData == 0:
            if idx+1 >= len(pe.sections):
                continue
            # fill the gap with null bytes..
            nsec = pe.sections[idx+1]
            nbase = nsec.VirtualAddress + baseaddr

            plen = nbase - secbase
            readsize = sec.SizeOfRawData if sec.SizeOfRawData < sec.VirtualSize else sec.VirtualSize
            secoff = pe.rvaToOffset(secrva)
            secbytes = pe.readAtOffset(secoff, readsize)
            secbytes += b'\x00' * plen
            vw.addMemoryMap(secbase, mapflags, fname, secbytes)
            vw.addSegment(secbase, len(secbytes), secname, fname)

            # Mark dead data on resource and import data directories
            if sec.VirtualAddress in deadvas:
                vw.markDeadData(secbase, secbase+len(secbytes))

            #FIXME create a mask for this
            if not (chars & PE.IMAGE_SCN_CNT_CODE) and not (chars & PE.IMAGE_SCN_MEM_EXECUTE) and not (chars & PE.IMAGE_SCN_MEM_WRITE):
                vw.markDeadData(secbase, secbase+len(secbytes))
            continue

        # if SizeOfRawData is greater than VirtualSize we'll end up using VS in our read..
        if sec.SizeOfRawData < sec.VirtualSize:
            if sec.SizeOfRawData > pe.filesize:
                continue

        #plen = sec.VirtualSize - sec.SizeOfRawData

        try:
            # According to http://code.google.com/p/corkami/wiki/PE#section_table if SizeOfRawData is larger than VirtualSize, VS is used..
            readsize = sec.SizeOfRawData if sec.SizeOfRawData < sec.VirtualSize else sec.VirtualSize

            secoff = pe.rvaToOffset(secrva)
            secbytes = pe.readAtOffset(secoff, readsize, shortok=True)
            slen = len(secbytes)
            secbytes += b'\x00' * (sec.VirtualSize - slen)

            slen = vw.addMemoryMap(secbase, mapflags, fname, secbytes, align=filealign)
            vw.addSegment(secbase, slen, secname, fname)

            # Mark dead data on resource and import data directories
            if sec.VirtualAddress in deadvas:
                vw.markDeadData(secbase, secbase+len(secbytes))

            # FIXME create a mask for this
            if not (chars & PE.IMAGE_SCN_CNT_CODE) and not (chars & PE.IMAGE_SCN_MEM_EXECUTE) and not (chars & PE.IMAGE_SCN_MEM_WRITE):
                vw.markDeadData(secbase, secbase+len(secbytes))

        except Exception as e:
            logger.warning("Error Loading Section (%s size:%d rva:%.8x offset: %d): %s", secname, secfsize, secrva, secoff, e)

    vw.addExport(entry, EXP_FUNCTION, '__entry', fname)
    vw.addEntryPoint(entry)

    # store the actual reloc section virtual address
    reloc_va = pe.getDataDirectory(PE.IMAGE_DIRECTORY_ENTRY_BASERELOC).VirtualAddress
    if reloc_va:
        reloc_va += baseaddr
    vw.setFileMeta(fname, "reloc_va", reloc_va)

    for rva, rtype in pe.getRelocations():

        # map PE reloc to VIV reloc ( or dont... )
        vtypedata = relmap.get(rtype)
        if vtypedata is None:
            logger.info('Skipping PE Relocation type: %d at %d (no handler)', rtype, rva)
            continue

        vtype, vtsize = vtypedata

        try:
            vtfmt = e_bits.getFormat(vtsize)    # for PowerPC, will need to get Endianness
            mapoffset = vw.readMemoryFormat(rva + baseaddr, vtfmt)[0] - imagebase
        except Exception as e:
            # the target address of the relocation is not accessible.
            # for example, it's not mapped, or split across sections, etc.
            # technically, the PE is corrupt.
            # by continuing on here, we are a bit more robust (but maybe incorrect)
            # than the Windows loader.
            #
            # discussed in:
            # https://github.com/vivisect/vivisect/issues/346
            logger.warning('Skipping invalid PE relocation: %d (%r)', rva, e)
            continue
        else:
            logger.info('PE relocation: 0x%x -> %r+0x%x', baseaddr+rva, fname, mapoffset)
            vw.addRelocation(rva + baseaddr, vtype, mapoffset)

    for rva, lname, iname in pe.getImports():
        if vw.probeMemory(rva + baseaddr, 4, e_const.MM_READ):
            vw.makeImport(rva + baseaddr, lname, iname)

    for rva, lname, iname in pe.getDelayImports():
        if vw.probeMemory(rva + baseaddr, 4, e_const.MM_READ):
            vw.makeImport(rva + baseaddr, lname, iname)
            if lname != '*':
                lname = vw.normFileName(lname)
            vw.setVaSetRow('DelayImports', (rva + baseaddr, lname + '.' + iname))

    # Tell vivisect about ntdll functions that don't exit...
    vw.addNoReturnApi("ntdll.RtlExitUserThread")
    vw.addNoReturnApi("kernel32.ExitProcess")
    vw.addNoReturnApi("kernel32.ExitThread")
    vw.addNoReturnApi("kernel32.FatalExit")
    vw.addNoReturnApiRegex(r"^msvcr.*\._CxxThrowException$")
    vw.addNoReturnApiRegex(r"^msvcr.*\.abort$")
    vw.addNoReturnApiRegex(r"^msvcr.*\.exit$")
    vw.addNoReturnApiRegex(r"^msvcr.*\._exit$")
    vw.addNoReturnApiRegex(r"^msvcr.*\.quick_exit$")
    vw.addNoReturnApiRegex(r"^api_ms_win_crt_runtime_.*\._invalid_parameter_noinfo_noreturn$")
    vw.addNoReturnApiRegex(r"^api_ms_win_crt_runtime_.*\.exit$")
    vw.addNoReturnApiRegex(r"^api_ms_win_crt_runtime_.*\._exit$")
    # https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/invalid-parameter-functions?view=vs-2019
    # TODO: Again, there's a couple in there that have conditional termination that we should check for
    #vw.addNoReturnApiRegex("vcruntime140.__std_terminate")
    # TODO: we should add abort and terminate on the conditions that there are no signal handlers
    # registered
    # https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/cexit-c-exit?view=vs-2019
    # vw.addNoReturnApiRegex("^msvcr.*\._cexit$")
    # vw.addNoReturnApiRegex("^msvcr.*\._c_exit$")
    vw.addNoReturnApi("ntoskrnl.KeBugCheckEx")


    exports = pe.getExports()
    for rva, ord, name in exports:
        eva = rva + baseaddr

        # Functions exported by ordinal only have no name
        if not name:
            name = "Ordinal_" + str(ord)

        try:
            vw.setVaSetRow('pe:ordinals', (eva, ord))
            vw.addExport(eva, EXP_UNTYPED, name, fname)
            if vw.probeMemory(eva, 1, e_const.MM_EXEC):
                vw.addEntryPoint(eva)
        except Exception as e:
            vw.vprint('addExport Failed: %s.%s (0x%.8x): %s' % (fname, name, eva, e))

    # Save off the ordinals...
    vw.setFileMeta(fname, 'ordinals', exports)

    fwds = pe.getForwarders()
    for rva, name, forwardname in fwds:
        vw.makeName(rva+baseaddr, "forwarder_%s.%s" % (fname, name))
        vw.makeString(rva+baseaddr)

    vw.setFileMeta(fname, 'forwarders', fwds)

    # Check For SafeSEH list...
    if pe.IMAGE_LOAD_CONFIG is not None:

        vw.setFileMeta(fname, "SafeSEH", True)

        va = pe.IMAGE_LOAD_CONFIG.SEHandlerTable
        if va != 0:
            vw.makeName(va, "%s.SEHandlerTable" % fname)
            count = pe.IMAGE_LOAD_CONFIG.SEHandlerCount
            # RP BUG FIX - sanity check the count
            if count * 4 < pe.filesize and vw.isValidPointer(va):
                # XXX - CHEAP HACK for some reason we have binaries still thorwing issues.. 

                try:
                    # Just cheat and use the workspace with memory maps in it already
                    for h in vw.readMemoryFormat(va, "<%dP" % count):
                        sehva = baseaddr + h
                        vw.addEntryPoint(sehva)
                        #vw.hintFunction(sehva, meta={'SafeSEH':True})
                except:
                    vw.vprint("SEHandlerTable parse error")

    # Last but not least, see if we have symbol support and use it if we do
    if vt_win32.dbghelp and filename:

        s = vt_win32.Win32SymbolParser(-1, filename, baseaddr)

        # We don't want exports or whatever because we already have them
        s.symopts |= vt_win32.SYMOPT_EXACT_SYMBOLS
        s.parse()

        # Add names for any symbols which are missing them
        for symname, symva, size, flags in s.symbols:

            if not vw.isValidPointer(symva):
                continue

            try:

                if vw.getName(symva) is None:
                    vw.makeName(symva, symname, filelocal=True)

            except Exception as e:
                vw.vprint("Symbol Load Error: %s" % e)

        # Also, lets set the locals/args name hints if we found any
        vw.setFileMeta(fname, 'PELocalHints', s._sym_locals)

    # if it has an EXCEPTION directory parse if it has the pdata
    edir = pe.getDataDirectory(PE.IMAGE_DIRECTORY_ENTRY_EXCEPTION)
    if edir.VirtualAddress and arch == 'amd64':
        va = edir.VirtualAddress + baseaddr
        vamax = va + edir.Size
        while va < vamax:
            try:
                f = vw.makeStructure(va, 'pe.IMAGE_RUNTIME_FUNCTION_ENTRY')
            except e_exc.SegmentationViolation as e:
                logger.warning('Invalid exception entry at 0x%x (error: %s)' % (va, str(e)))
                break

            if not vw.isValidPointer(baseaddr + f.UnwindInfoAddress):
                break

            # FIXME UNWIND_INFO *requires* DWORD alignment, how is it enforced?
            fva = f.BeginAddress + baseaddr
            uiva = baseaddr + f.UnwindInfoAddress
            # Possible method 1...
            #uiva = baseaddr + (f.UnwindInfoAddress & 0xfffffffc )

            # Possible method 2...
            #uirem = f.UnwindInfoAddress % 4
            #if uirem:
                #uiva += ( 4 - uirem )
            uinfo = vw.getStructure(uiva, 'pe.UNWIND_INFO')
            ver = uinfo.VerFlags & 0x7
            if ver != 1:
                vw.vprint('Unwind Info Version: %d (bailing on .pdata)' % ver)
                break

            flags = uinfo.VerFlags >> 3
            # Check if it's a function *block* rather than a function *entry*
            if not (flags & PE.UNW_FLAG_CHAININFO):
                vw.addEntryPoint(fva)

            va += len(f)

    # auto-mark embedded PEs as "dead data" to prevent code flow...
    if carvepes:
        pe.fd.seek(0)
        fbytes = pe.fd.read()
        for offset, i in pe_carve.carve(fbytes, 1):
            try:
                # ensure sub-pe can be parsed
                subpe = pe_carve.CarvedPE(fbytes, offset, [i])
            except vivisect.exc.CorruptPeFile:
                logger.warning("could not parse carved PE at offset 0x%x, XOR key: 0x%x", offset, i)
                continue

            pebytes = subpe.readAtOffset(0, subpe.getFileSize())
            rva = pe.offsetToRva(offset) + baseaddr
            vw.markDeadData(rva, rva+len(pebytes))

    return fname

def getMemBaseAndSize(vw, pe, baseaddr=None):
    '''
    Returns the default baseaddr and memory size required to load the file
    '''
    savebase = baseaddr

    if baseaddr is None:
        baseaddr = pe.IMAGE_NT_HEADERS.OptionalHeader.ImageBase

    memmaps = [(baseaddr, e_const.MM_READ, '', 0x1000)]
    codesize = pe.IMAGE_NT_HEADERS.OptionalHeader.SizeOfCode
    for idx, sec in enumerate(pe.sections):
        secrva = sec.VirtualAddress
        secvsize = sec.VirtualSize
        secfsize = sec.SizeOfRawData
        secbase = secrva + baseaddr
        secname = sec.Name.strip("\x00")
        secrvamax = secrva + secvsize

        if sec.VirtualSize == 0 or sec.SizeOfRawData == 0:
            if idx+1 >= len(pe.sections):
                continue
            # fill the gap with null bytes..
            nsec = pe.sections[idx+1]
            nbase = nsec.VirtualAddress + baseaddr

            mlen = nbase - secbase
            memmaps.append((secbase, 7, '', mlen))
            continue

        if sec.SizeOfRawData < sec.VirtualSize:
            if sec.SizeOfRawData > pe.filesize:
                continue

        mlen = sec.VirtualSize
        memmaps.append((secbase, 0, '', mlen))

    baseaddr = 0xffffffffffffffffffffffff
    topmem = 0

    for mapva, mperms, mname, lenbytes in memmaps:
        if mapva < baseaddr:
            baseaddr = mapva
        endva = mapva + lenbytes
        if endva > topmem:
            topmem = endva

    size = topmem - baseaddr

    if baseaddr == 0:
        baseaddr = vw.config.viv.parsers.pe.baseaddr

    if savebase:
        # if we provided a baseaddr, override what the file wants
        baseaddr = savebase
        
    return baseaddr, size




'''
Utilities for windows tracer objects.
'''
import PE
import vtrace


def deAslr(trace, va):
    '''
    Given an address in an ASLR'd library, rebase
    it back to the address as it would be if the
    given PE were at it's suggested address...
    '''

    if vtrace.remote:
        raise Exception('deAslr only works for local debuggers!')

    map = trace.getMemoryMap(va)
    if map is None:
        return va

    mapva, mapsize, mapperm, mapfname = map
    if not mapfname:
        return va

    normname = trace.normFileName(mapfname)
    sym = trace.getSymByName(normname)
    if sym is None:
        return va

    membase = int(sym)

    pe = PE.peFromFileName(mapfname)
    filebase = pe.IMAGE_NT_HEADERS.OptionalHeader.ImageBase

    rva = va - membase

    return filebase + rva



import logging

import envi.bits as e_bits

logger = logging.getLogger(__name__)


def ffTermFptrArray(vw, va, max=100):
    ret = []
    ffterm = e_bits.u_maxes[vw.psize]
    for i in range(max):

        vw.makeNumber(va, vw.psize)

        val = vw.parseNumber(va, vw.psize)
        if val == ffterm:
            return ret

        try:
            logger.debug('ffTermFptrArray(): discovered new function: 0x%x', val)
            vw.makeFunction(val)
            ret.append(val)
        except Exception as e:
            logger.warning("FIXME (ffTermFptrArray): %s", e)
        va += vw.psize
    return ret


def analyze(vw):

    # Go through the elf sections and handle known types.
    for segva, segsize, segname, segfname in vw.getSegments():
        if segname == ".ctors":
            if vw.getLocation(segva) is not None:  # Check if it's already done
                continue
            for f in ffTermFptrArray(vw, segva):
                vw.makeName(f, "ctor_%.8x" % f)

        elif segname == ".dtors":
            if vw.getLocation(segva) is not None:  # Check if it's already done
                continue
            for f in ffTermFptrArray(vw, segva):
                vw.makeName(f, "dtor_%.8x" % f)

        elif segname == ".plt":
            pass
            # FIXME: Do linear disassembly of the PLT here...



"""
Cobra RMI Framework

Cobra is a remote method invocation interface that is very "pythony".  It is
MUCH like its inspiration pyro, but slimmer and safer for things like threading
and object de-registration.  Essentially, cobra allows you to call methods from
and get/set attributes on objects that exist on a remote system.

"""
# Copyright (C) 2011 Invisigoth - See LICENSE file for details
import os
import json
import time
import types
import queue
import pickle
import socket
import struct
import logging
import traceback
import urllib.parse
import envi.common as e_cmn
import envi.config as e_config

from threading import current_thread, Thread, RLock, Timer, Lock
from socketserver import ThreadingTCPServer, BaseRequestHandler
try:
    import msgpack
    dumpargs = {}
    loadargs = {'use_list': 0}
    if msgpack.version >= (0, 4, 1):
        dumpargs['use_bin_type'] = 1
        if msgpack.version < (1, 0, 0):
            loadargs['encoding'] = 'utf-8'
        else:
            loadargs['strict_map_key'] = False

except ImportError:
    msgpack = None

logger = logging.getLogger(__name__)

defconfig = {

    'cobra':{
        'base_timeout': 120,
    },

}

docconfig = {
    'cobra':{
        'base_timeout': "Default timeout for cobra connections",
    }
}

daemon = None
version = "Cobra2"
COBRA_PORT=5656
COBRASSL_PORT=5653
cobra_retrymax = None # Optional *global* retry max count

socket_builders = {}    # Registered socket builders

# Message Types
COBRA_HELLO     = 0
COBRA_CALL      = 1
COBRA_GETATTR   = 2
COBRA_SETATTR   = 3
COBRA_ERROR     = 4
COBRA_GOODBYE   = 5
COBRA_AUTH      = 6
COBRA_NEWOBJ    = 7 # Used to return object references

SFLAG_MSGPACK   = 0x0001
SFLAG_JSON      = 0x0002

class CobraException(Exception):
    """Base for Cobra exceptions"""
    pass

class CobraClosedException(CobraException):
    """Raised when a connection is unexpectedly closed."""
    pass

class CobraRetryException(CobraException):
    """Raised when the retrymax (if present) for a proxy object is exceeded."""
    pass

class CobraPickleException(CobraException):
    """Raised when pickling fails."""
    pass

class CobraAuthException(CobraException):
    '''Raised when specified auth data is rejected'''
    pass

class CobraPermDenied(CobraException):
    '''Raised when a call/setattr/getattr is not allowed'''

class CobraErrorException(Exception):
    '''
    Raised when we receive a COBRA_ERROR message and the current options
    dont support serializing exception objects.
    '''

def connectSocket(host, port, timeout=None):
    """
    Make the long names go away....
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if timeout is not None:
        s.settimeout(timeout)

    s.connect((host, port))

    return s

def getCallerInfo():
    """
    This function may be used from *inside* a method being called
    by a remote caller.  It will return a tuple of host,port for the
    other side of the connection... use wisely ;)
    """
    return getattr(current_thread(), "_cobra_caller_info", None)

def getLocalInfo():
    """
    This function returns the local host,port combination being
    used in the socket servicing the current request
    """
    return getattr(current_thread(), "_cobra_local_info", None)

def getUserInfo():
    '''
    Get the cobra authenticated username of the current user
    ( or None if no user was authenticated )
    '''
    return getattr(current_thread(), "_cobra_authuser", None)

def setCallerInfo(callerinfo):
    """
    This is necessary because of crazy python method call
    name munging for thread attributes ;)
    """
    current_thread()._cobra_caller_info = callerinfo

def setUserInfo(authuser):
    current_thread()._cobra_authuser = authuser

def setLocalInfo(localinfo):
    current_thread()._cobra_local_info = localinfo

def nocobra(f):
    f.__no_cobra__ = True
    return f

def newobj(f):
    f._cobra_newobj = True
    return f

def newobjwith(f):
    f._cobra_newobj = True
    f._cobra_newobjwith = True
    return f

class CobraMethod:
    def __init__(self, proxy, methname):
        self.proxy = proxy
        self.methname = methname
        self.__name__ = methname

    def __call__(self, *args, **kwargs):
        name = self.proxy._cobra_name
        logger.debug("Calling: %s, %s, %s, %s", name, self.methname, repr(args)[:20], repr(kwargs)[:20]) 
        casync = kwargs.pop('_cobra_async', None)
        if casync:
            csock = self.proxy._cobra_getsock()
            return csock.cobraAsyncTransaction(COBRA_CALL, name, (self.methname, args, kwargs))

        with self.proxy._cobra_getsock() as csock:
            mtype, name, data = csock.cobraTransaction(COBRA_CALL, name, (self.methname, args, kwargs))

        if mtype == COBRA_CALL:
            return data
        if mtype == COBRA_NEWOBJ:
            uri = swapCobraObject(self.proxy._cobra_uri, data)
            return CobraProxy(uri)
        raise data

def pickledumps(o):
    return pickle.dumps( o, protocol=pickle.HIGHEST_PROTOCOL )

def jsonloads(b):
    return json.loads(b)

def jsondumps(b):
    return json.dumps(b)

class CobraSocket:

    def __init__(self, socket, sflags=0):
        self.sflags = sflags
        self.socket = socket
        self.dumps = pickledumps
        self.loads = pickle.loads

        if sflags & SFLAG_MSGPACK:
            if not msgpack:
                raise Exception('Missing "msgpack" python module ( http://visi.kenshoto.com/viki/Msgpack )')

            def msgpackloads(b):
                logger.log(e_cmn.MIRE, "<< %r  (%r)", b, loadargs)
                return msgpack.loads(b, **loadargs)

            def msgpackdumps(b):
                logger.log(e_cmn.MIRE, ">> %r  (%r)", b, dumpargs)
                return msgpack.dumps(b, **dumpargs)

            self.dumps = msgpackdumps
            self.loads = msgpackloads

        if sflags & SFLAG_JSON:
            self.dumps = jsondumps
            self.loads = jsonloads

    def __del__(self):
        self.socket.close()

    def getSockName(self):
        return self.socket.getsockname()

    def getPeerName(self):
        return self.socket.getpeername()

    def sendMessage(self, mtype, objname, data):
        """
        Send message is responsable for transmission of cobra messages,
        and socket reconnection in the event that the send fails for network
        reasons.
        """

        # NOTE: for errors while using msgpack, we must send only the str
        if mtype == COBRA_ERROR and self.sflags & (SFLAG_MSGPACK | SFLAG_JSON):
            data = str(data)

        try:
            buf = self.dumps(data)
        except Exception as e:
            raise CobraPickleException("The arguments/attributes must be serializable: %s" % e)

        obj = objname.encode('utf-8')
        self.sendExact(struct.pack("<III", mtype, len(obj), len(buf)) + obj + buf)

    def recvMessage(self):
        """
        Returns tuple of mtype, objname, and data
        This method is *NOT* responsable for re-connection, because there
        is not context on the server side for what to send on re-connect.
        Client side uses of the CobraSocket object should use cobraTransaction
        to ensure re-tranmission of the request on reception errors.
        """
        hdr = self.recvExact(12)
        mtype, nsize, dsize = struct.unpack("<III", hdr)
        name = self.recvExact(nsize).decode('utf-8')
        data = self.loads(self.recvExact(dsize))

        # NOTE: for errors while using msgpack, we must send only the str
        if mtype == COBRA_ERROR and self.sflags & (SFLAG_MSGPACK | SFLAG_JSON):
            data = CobraErrorException(data)
        return (mtype, name, data)

    def recvExact(self, size):
        buf = b""
        s = self.socket
        while len(buf) != size:
            x = s.recv(size - len(buf))
            if len(x) == 0:
                raise CobraClosedException("Socket closed in recvExact...")
            buf += x
        return buf

    def sendExact(self, buf):
        self.socket.sendall(buf)

class SocketBuilder:

    def __init__(self, host, port, timeout=None):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.retrymax = None

        self.ssl = False
        self.sslca = None
        self.sslcrt = None
        self.sslkey = None

    def setTimeout(self, timeout):
        '''
        Set the timeout for newly created sockets.
        '''
        self.timeout = timeout

    def setSslEnabled(self, status):
        self.ssl = status

    def setSslCa(self, crtfile):
        '''
        Set the SSL Certificate Authority for this socket builder.

        ( This enables checking the server's presented cert )
        '''
        self.ssl = True
        self.sslca = crtfile

    def setSslClientCert(self, crtfile, keyfile):
        '''
        Set the cert/key used by this client to negotiate SSL.
        '''
        self.ssl = True
        self.sslcrt = crtfile
        self.sslkey = keyfile

    def __call__(self):

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.timeout is not None:
            sock.settimeout(self.timeout)

        if self.ssl:
            import ssl
            sslkwargs = {}

            if self.sslca:
                sslkwargs['ca_certs'] = self.sslca
                sslkwargs['cert_reqs']=ssl.CERT_REQUIRED

            if self.sslcrt and self.sslkey:
                sslkwargs['keyfile'] = self.sslkey
                sslkwargs['certfile'] = self.sslcrt

            sock = ssl.wrap_socket(sock, **sslkwargs)

        sock.connect((self.host, self.port))
        return sock

class CobraAsyncTrans:

    def __init__(self, csock, mtype, objname, data):
        self.data = data
        self.csock = csock
        self.mtype = mtype
        self.objname = objname
  
        # Issue the call..
        self.asyncCobraTransaction()

    def asyncCobraTransaction(self):
        """
        This is an API for clients to use.  It will retransmit
        a sendMessage() automagically on recpt of an exception
        in recvMessage()
        """
        while True:
            try:
                self.csock.sendMessage(self.mtype, self.objname, self.data)
                return
            except CobraAuthException as e:
                raise
            except (socket.error,CobraClosedException):
                self.csock.reConnect()

    def wait(self):
        try:
            while True:
                try:
                    mtype,name,data = self.csock.recvMessage()
                    if mtype == COBRA_CALL:
                        return data
                    raise data

                except CobraAuthException as e:
                    raise

                except (socket.error, CobraClosedException) as e:
                    # force a reconnect
                    self.csock.reConnect()
                    self.asyncCobraTransaction()
        finally:
            if self.csock.pool:
                self.csock.pool.put(self.csock)
            self.csock = None

class CobraClientSocket(CobraSocket):

    def __init__(self, sockctor, retrymax=cobra_retrymax, sflags=0, authinfo=None, pool=None):
        CobraSocket.__init__(self, sockctor(), sflags=sflags)
        self.sockctor = sockctor
        self.retries = 0
        self.trashed = False
        self.retrymax = retrymax
        self.authinfo = authinfo
        self.pool     = pool

    def __enter__(self):
        return self 

    def __exit__(self, extype, value, tb):
        if self.pool:
            self.pool.put(self) 

    def reConnect(self):
        """
        Handle the event where we need to reconnect
        """
        while self.retrymax is None or self.retries < self.retrymax:
            logger.info("COBRA: Reconnection Attempt\n")
            try:
                if self.socket:
                    self.socket.close()

                self.socket = self.sockctor()

                # A bit messy but... a fix for now...
                # If we have authinfo lets authenticate
                authinfo = self.authinfo
                if authinfo is not None:
                    self.sendMessage(COBRA_AUTH, '', authinfo)
                    mtype,rver,data = self.recvMessage()
                    if mtype != COBRA_AUTH:
                        raise CobraAuthException('Authentication Failed!')

                self.retries = 0
                return

            except CobraAuthException as e:
                raise

            except Exception as e:
                logger.warning("reConnect hit exception: %s" % str(e))
                time.sleep( max(2 ** self.retries, 10) )
                self.retries += 1

        self.trashed = True
        raise CobraRetryException()

    def cobraAsyncTransaction(self, mtype, objname, data):
        return CobraAsyncTrans(self, mtype, objname, data)

    def cobraTransaction(self, mtype, objname, data):
        """
        This is an API for clients to use.  It will retransmit
        a sendMessage() automagically on recpt of an exception
        in recvMessage()
        """
        while True:
            try:
                self.sendMessage(mtype, objname, data)
                return self.recvMessage()

            except CobraAuthException:
                raise

            except CobraClosedException:
                self.reConnect()

            except socket.error:
                self.reConnect()

class CobraDaemon(ThreadingTCPServer):

    def __init__(self, host="", port=COBRA_PORT, sslcrt=None, sslkey=None, sslca=None, msgpack=False, json=False, config=None):
        '''
        Construct a cobra daemon object.

        Parameters:
        host        - Optional hostname/ip to bind the service to (default: inaddr_any)
        port        - The port to bind (Default: COBRA_PORT)
        msgpack     - Use msgpack serialization

        # SSL Options
        sslcrt / sslkey     - Specify sslcrt and sslkey to enable SSL server side
        sslca               - Specify an SSL CA key to use validating client certs

        '''
        if config is None:
            # If we don't provide a config, grab a default
            config = e_config.loadConfig(defconfig, docconfig)

        self.config = config
        self.thr = None
        self.run = True
        self.shared = {}
        self.dowith = {}
        self.host = host
        self.port = port
        self.reflock = RLock()
        self.refcnts = {}
        self.authmod = None
        self.sflags = 0

        self.allow_reuse_address = True

        if msgpack and json:
            raise Exception('CobraDaemon can not use both msgpack *and* json!')

        if msgpack:
            requireMsgpack()
            self.sflags |= SFLAG_MSGPACK

        if json:
            self.sflags |= SFLAG_JSON

        # SSL Options
        self.sslca = sslca
        self.sslcrt = sslcrt
        self.sslkey = sslkey

        self.cansetattr = True
        self.cangetattr = True

        if sslcrt and not os.path.isfile(sslcrt):
            raise Exception('CobraDaemon: sslcrt param must be a file!')

        if sslkey and not os.path.isfile(sslkey):
            raise Exception('CobraDaemon: sslkey param must be a file!')

        if sslca and not os.path.isfile(sslca):
            raise Exception('CobraDaemon: sslca param must be a file!')

        self.allow_reuse_address = True
        ThreadingTCPServer.__init__(self, (host, port), CobraRequestHandler)

        if port == 0:
            self.port = self.socket.getsockname()[1]

        self.daemon_threads = True
        self.recvtimeout = None

    def logCallerError(self, oname, args, msg=""):
        pass

    def setGetAttrEnabled(self, status):
        self.cangetattr = status

    def setSetAttrEnabled(self, status):
        self.cansetattr = status

    def setSslCa(self, crtfile):
        '''
        Set the SSL Certificate Authority by this server.
        ( to validate client certs )
        '''
        self.sslca = crtfile

    def setSslServerCert(self, crtfile, keyfile):
        '''
        Set the cert/key used by this server to negotiate SSL.
        '''
        self.sslcrt = crtfile
        self.sslkey = keyfile

    def fireThread(self):
        self.thr = Thread(target=self.serve_forever, daemon=True)
        self.thr.start()

    def stopServer(self):
        self.run = False
        self.shutdown()
        self.server_close()
        self.thr.join()

    def serve_forever(self):
        try:

            ThreadingTCPServer.serve_forever(self)

        except Exception as e:
            if not self.run:
                return

            raise

    def setAuthModule(self, authmod):
        '''
        Enable an authentication module for this server
        ( all connections *must* be authenticated through the authmod )

        NOTE: See cobra.auth.* for various auth module implementations

        Example:
            import cobra.auth.shadow as c_a_shadow
            authmod = c_a_shadow.ShadowFileAuth('passwdfile.txt')
            cdaemon = CobraDaemon()
            cdaemon.setAuthModule()
        '''
        self.authmod = authmod

    def getSharedObject(self, name):
        return self.shared.get(name, None)

    def getSharedObjects(self):
        '''
        Return a list of (name, obj) for the currently shared objects.

        Example:
            for name,obj in daemon.getSharedObjects():
                print('%s: %r' % (name,obj))
        '''
        return self.shared.items()

    def getSharedName(self, obj):
        '''
        If this object is shared already, get the name...
        '''
        for name, sobj in self.shared.items():
            if sobj == obj:
                return name
        return None

    def getRandomName(self):
        return ''.join(['%.2x' % x for x in os.urandom(16)])

    def shareObject(self, obj, name=None, doref=False, dowith=False):
        """
        Share an object in this cobra server.  By specifying
        doref=True you will let CobraProxy objects decide that
        the object is done and should be un-shared.  Also, if
        name is None a random name is chosen.  Use dowith=True
        to cause sharing/unsharing to enter/exit (requires doref=True).

        Returns: name (or the newly generated random one)

        """
        refcnt = None
        if dowith and not doref:
            raise Exception('dowith *requires* doref!')

        if doref:
            refcnt = 0

        if dowith:
            obj.__enter__()

        if name is None:
            name = self.getRandomName()

        self.shared[name] = obj
        self.dowith[name] = dowith
        self.refcnts[name] = refcnt
        return name

    def getObjectRefCount(self, name):
        return self.refcnts.get(name)

    def decrefObject(self, name, ok=True):
        """
        Decref this object and if it reaches 0, unshare it.
        """
        logger.debug('Decrementing: %s', name)
        self.reflock.acquire()
        try:

            refcnt = self.refcnts.get(name, None)
            if refcnt is not None:
                refcnt -= 1
                self.refcnts[name] = refcnt
                if refcnt == 0:
                    self.unshareObject(name,ok=ok)

        finally:
            self.reflock.release()

    def increfObject(self, name):
        logger.debug('Incrementing: %s', name)
        self.reflock.acquire()
        try:
            refcnt = self.refcnts.get(name, None)
            if refcnt is not None:
                refcnt += 1
                self.refcnts[name] = refcnt
        finally:
            self.reflock.release()

    def unshareObject(self, name, ok=True):
        logger.debug('Unsharing %s', name)
        self.refcnts.pop(name, None)
        obj = self.shared.pop(name, None)

        # If we are using a with block, notify it
        if self.dowith.pop(name, False):
            args = (None,None,None)
            if not ok:
                args = (Exception, Exception('with boom'), None)
            obj.__exit__(*args)
        return obj

class CobraRequestHandler(BaseRequestHandler):

    def handle(self):
        c = CobraConnectionHandler(self.server, self.request)
        c.handleClient()

class CobraConnectionHandler:

    def __init__(self, daemon, socket):
        self.daemon = daemon
        self.socket = socket
        self.handlers = (
            self.handleHello,
            self.handleCall,
            self.handleGetAttr,
            self.handleSetAttr,
            self.handleError,
            self.handleGoodbye,
            self.handleError,
        )

    def handleClient(self):

        peer = self.socket.getpeername()
        me = self.socket.getsockname()
        logger.info("Got a connection from: %s" % str(peer))

        sock = self.socket
        if self.daemon.sslkey:
            import ssl
            sslca = self.daemon.sslca
            keyfile = self.daemon.sslkey
            certfile = self.daemon.sslcrt
            sslreq = ssl.CERT_NONE
            # If they specify a CA key, require valid client certs
            if sslca:
                sslreq=ssl.CERT_REQUIRED

            sock = ssl.wrap_socket(sock,
                                     keyfile=keyfile, certfile=certfile,
                                     ca_certs=sslca, cert_reqs=sslreq,
                                     server_side=True)

        if self.daemon.recvtimeout:
            sock.settimeout( self.daemon.recvtimeout )

        authuser = None

        csock = CobraSocket(sock, sflags=self.daemon.sflags)

        setCallerInfo(peer)
        setLocalInfo(me)

        # If we have an authmod, they must send an auth message first
        if self.daemon.authmod:
            mtype,name,data = csock.recvMessage()
            if mtype != COBRA_AUTH:
                csock.sendMessage(COBRA_ERROR, '', CobraAuthException('Authentication Required!'))
                return

            authuser = self.daemon.authmod.authCobraUser( data )
            if not authuser:
                csock.sendMessage(COBRA_ERROR, '', CobraAuthException('Authentication Failed!'))
                return

            csock.sendMessage(COBRA_AUTH, '', authuser)
            setUserInfo( authuser )

        while True:

            try:
                mtype,name,data = csock.recvMessage()
            except CobraClosedException:
                break
            except socket.error as e:
                logger.warning("Cobra socket error in handleClient. Err: %s", str(e))
                break

            # If they re-auth ( app layer ) later, lets handle it...
            if mtype == COBRA_AUTH and self.daemon.authmod:
                authuser = self.daemon.authmod.authCobraUser(data)
                if not authuser:
                    csock.sendMessage(COBRA_ERROR,'',CobraAuthException('Authentication Failed!'))
                    continue

                setUserInfo(authuser)
                csock.sendMessage(COBRA_AUTH, '', authuser)
                continue

            if self.daemon.authmod and not self.daemon.authmod.checkUserAccess( authuser, name ):
                csock.sendMessage(COBRA_ERROR, name, Exception('Access Denied For User: %s' % authuser))
                continue

            obj = self.daemon.getSharedObject(name)
            logger.debug("MSG FOR: %s:%s", str(name), type(obj))
            if obj is None:
                try:
                    csock.sendMessage(COBRA_ERROR, name, Exception("Unknown object requested: %s" % name))
                except CobraClosedException:
                    pass
                logger.warning("Got request for unknown object: %s" % name)
                continue

            try:
                handler = self.handlers[mtype]
            except:
                try:
                    csock.sendMessage(COBRA_ERROR, name, Exception("Invalid Message Type"))
                except CobraClosedException:
                    pass
                logger.warning("Got Invalid Message Type: %d for %s" % (mtype, data))
                continue

            try:
                handler(csock, name, obj, data)
            except Exception as e:
                logger.warning("cobra handler hit exception: %s" % str(e))
                logger.debug("", exc_info=1)
                try:
                    csock.sendMessage(COBRA_ERROR, name, e)
                except TypeError as typee:
                    # Probably about pickling...
                    csock.sendMessage(COBRA_ERROR, name, Exception(str(e)))
                except CobraClosedException:
                    pass

    def handleError(self, csock, oname, obj, data):
        raise NotImplementedError("How did we hit handleError?")

    def handleHello(self, csock, oname, obj, data):
        """
        Hello messages are used to get the initial cache of
        method names for the newly connected object.
        """
        logger.debug("Hello")
        self.daemon.increfObject(oname)
        ret = {}
        for name in dir(obj):
            attr = getattr(obj, name, None)
            if isinstance(attr, (types.MethodType, types.BuiltinMethodType, types.FunctionType, CobraMethod)):
                ret[name] = True
        try:
            csock.sendMessage(COBRA_HELLO, version, ret)
        except CobraClosedException:
            pass

    def handleCall(self, csock, oname, obj, data):
        logger.debug("Calling %s", str(data))
        methodname, args, kwargs = data
        meth = getattr(obj, methodname)
        if getattr(meth,'__no_cobra__',False):
            raise CobraPermDenied('%s is tagged nocall!' % methodname)
        try:
            ret = meth(*args, **kwargs)

            if getattr(meth,'_cobra_newobj',None):
                dowith = getattr(meth,'_cobra_newobjwith',False)
                objname = self.daemon.shareObject(ret, doref=True, dowith=dowith)
                csock.sendMessage(COBRA_NEWOBJ, "", objname)
                return

            csock.sendMessage(COBRA_CALL, "", ret)

        except CobraClosedException:
            pass
        except Exception as e:
            self.daemon.logCallerError(oname, data, msg=traceback.format_exc())
            raise

    def handleGetAttr(self, csock, oname, obj, name):
        logger.debug("Getting Attribute: %s", str(name))
        if not self.daemon.cangetattr:
            raise CobraPermDenied('getattr disallowed!')
        try:
            csock.sendMessage(COBRA_GETATTR, "", getattr(obj, name))
        except CobraClosedException:
            pass

    def handleSetAttr(self, csock, oname, obj, data):
        logger.debug("Setting Attribute: %s", str(data))
        if not self.daemon.cansetattr:
            raise CobraPermDenied('setattr disallowed!')
        name,value = data
        setattr(obj, name, value)
        try:
            csock.sendMessage(COBRA_SETATTR, "", "")
        except CobraClosedException:
            pass

    def handleGoodbye(self, csock, oname, obj, data):
        logger.debug("Goodbye")
        self.daemon.decrefObject(oname,ok=data)
        try:
            csock.sendMessage(COBRA_GOODBYE, "", "")
        except CobraClosedException:
            pass

def isCobraUri(uri):
    try:
        x = urllib.parse.urlparse(uri)
        if x.scheme not in ["cobra", "cobrassl"]:
            return False
    except Exception as e:
        return False
    return True

def chopCobraUri(uri):
    purl = urllib.parse.urlparse(uri)
    scheme = purl.scheme
    host = purl.hostname
    name = purl.path.strip('/')

    port = purl.port
    if not port:
        port = COBRA_PORT

    # Do we have any URL options?
    urlparams = {}
    for urlopt in purl.query.split('&'):
        urlval = 1
        if urlopt.find('=') != -1:
            urlopt,urlval = urlopt.split('=',1)

        urlopt = urlopt.lower()
        urlparams[urlopt] = urlval

    return scheme,host,port,name,urlparams

class CobraProxy:
    '''
    A proxy object for remote objects shared with Cobra

    A few optional keyword arguments are handled by all cobra protocols:
        retrymax    - Max transparent reconnect attempts
        timeout     - Socket timeout for a cobra socket
        authinfo    - A dict, probably like {'user':'username','passwd':'mypass'}
                      ( but it can be auth module specific )
        msgpack     - Use msgpack serialization
        sockpool    - Fixed sized pool of cobra sockets (not socket per thread)

    Also, the following protocol options may be passed through the URI:

    msgpack=1
    authinfo=<base64( json( <authinfo dict> ))>
    '''

    def __init__(self, URI, retrymax=None, timeout=None, **kwargs):

        scheme, host, port, name, urlparams = chopCobraUri( URI )

        logger.debug("Spinning up CobraProxy on %s:%s with object: %s", host, port, repr(name))

        self._cobra_uri = URI
        self._cobra_scheme = scheme
        self._cobra_host = host
        self._cobra_port = port
        self._cobra_slookup = (host,port)
        self._cobra_name = name
        self._cobra_retrymax = urlparams.get('retrymax', retrymax)
        self._cobra_timeout = urlparams.get('timeout', timeout)
        self._cobra_kwargs = kwargs
        self._cobra_gothello = False
        self._cobra_sflags = 0
        self._cobra_spoolcnt = int(urlparams.get('sockpool', 0))
        self._cobra_sockpool = None

        self._cobra_config = kwargs.get('config')
        if self._cobra_config is None:
            # If we don't provide a config, grab a default
            self._cobra_config = e_config.loadConfig(defconfig, docconfig)


        if self._cobra_timeout is not None:
            self._cobra_timeout = int(self._cobra_timeout)

        if self._cobra_retrymax is not None:
            self._cobra_retrymax = int(self._cobra_retrymax)

        if urlparams.get('msgpack'):
            requireMsgpack()
            self._cobra_sflags |= SFLAG_MSGPACK

        if urlparams.get('json'):
            self._cobra_sflags |= SFLAG_JSON

        urlauth = urlparams.get('authinfo')
        if urlauth:
            authinfo = json.loads(urlauth.decode('base64'))
            self._cobra_kwargs['authinfo'] = authinfo

        # If they asked for msgpack
        if kwargs.get('msgpack'):
            requireMsgpack()
            self._cobra_sflags |= SFLAG_MSGPACK

        if kwargs.get('json'):
            self._cobra_sflags |= SFLAG_JSON

        if self._cobra_spoolcnt:
            self._cobra_sockpool = queue.Queue()
            # timeout reqeuired for pool usage
            if not self._cobra_timeout:
                self._cobra_timeout = self._cobra_config.cobra.base_timeout
            # retry max required on pooling
            if not self._cobra_retrymax:
                self._cobra_retrymax = 3

            [self._cobra_sockpool.put(self._cobra_newsock()) for i in range(self._cobra_spoolcnt)]

        # If we got passed as user/passwd in our kwargs
        with self._cobra_getsock() as csock:
            mtype,rver,data = csock.cobraTransaction(COBRA_HELLO, name, "")

            if mtype == COBRA_ERROR:
                csock.trashed = True
                if self._cobra_sflags & (SFLAG_MSGPACK|SFLAG_JSON):
                    data = Exception(data)
                raise data

            if rver != version:
                csock.trashed = True
                raise Exception("Server Version Not Supported: %s" % rver)

            if mtype != COBRA_HELLO:
                csock.trashed = True
                raise Exception("Invalid Cobra Hello Response")

        self._cobra_gothello = True
        self._cobra_methods = data

    def cobraAuthenticate(self, authinfo):
        '''
        Re-authenticate to the server ( and store auth info for reconnect ).
        '''
        with self._cobra_getsock() as csock:
            mtype,rver,data = csock.cobraTransaction(COBRA_AUTH, '', authinfo)
        if mtype == COBRA_AUTH:
            self._cobra_kwargs['authinfo'] = authinfo
            return True
        return False

    def _cobra_getsock(self, thr=None):
        if self._cobra_spoolcnt:
            sock = self._cobra_sockpool.get()
        else:
            if not thr: # if thread isn't specified, use the current thread
                thr = current_thread()

            tsocks = getattr(thr, 'cobrasocks', None)
            if tsocks is None:
                tsocks = {}
                thr.cobrasocks = tsocks
            sock = tsocks.get(self._cobra_slookup)

        if not sock or sock.trashed:
            # Lets build a new socket... shall we?
            sock = self._cobra_newsock()
            # If we have authinfo lets authenticate
            authinfo = self._cobra_kwargs.get('authinfo')
            if authinfo is not None:
                mtype,rver,data = sock.cobraTransaction(COBRA_AUTH, '', authinfo)
                if mtype != COBRA_AUTH:
                    raise CobraAuthException('Authentication Failed!')

            if not self._cobra_spoolcnt:
                tsocks[self._cobra_slookup] = sock
        return sock

    def _cobra_newsock(self):
        """
        This is only used by *clients*
        """
        host = self._cobra_host
        port = self._cobra_port
        timeout = self._cobra_timeout
        retrymax = self._cobra_retrymax

        builder = getSocketBuilder(host,port)
        if builder is None:
            builder = SocketBuilder(host,port)
            builder.setTimeout(timeout) # Might be None... 
            if self._cobra_scheme == 'cobrassl':
                builder.setSslEnabled(True)

            addSocketBuilder(host, port, builder)

        authinfo = self._cobra_kwargs.get('authinfo') 
        return CobraClientSocket(builder, retrymax=retrymax, sflags=self._cobra_sflags, authinfo=authinfo, pool=self._cobra_sockpool)

    def __dir__(self):
        '''
        return a list of proxied method names
        '''
        return self._cobra_methods.keys()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, sdict):
        self.__dict__.update(sdict)

    def __hash__(self):
        return hash(self._cobra_uri)

    def __nonzero__(self):
        return True

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "<CobraProxy %s>" % self._cobra_uri

    def __eq__(self, obj):
        ouri = getattr(obj, '_cobra_uri', None)
        return self._cobra_uri == ouri

    def __ne__(self, obj):
        if self == obj:
            return False
        return True

    def __setattr__(self, name, value):
        logger.debug('Setattr: %s:%s', name, repr(value)[:20])

        if name.startswith('_cobra_'):
            self.__dict__[name] = value
            return

        with self._cobra_getsock() as csock:
            mtype,name,data = csock.cobraTransaction(COBRA_SETATTR, self._cobra_name, (name, value))
        if mtype == COBRA_ERROR:
            raise data
        elif mtype == COBRA_SETATTR:
            return
        else:
            raise Exception("Invalid Cobra Response")

    def __getattr__(self, name):
        logger.debug('Getattr: %s', name)

        if name == "__getinitargs__":
            raise AttributeError()

        # Handle methods
        if self._cobra_methods.get(name, False):
            return CobraMethod(self, name)

        with self._cobra_getsock() as csock:
            mtype, name, data = csock.cobraTransaction(COBRA_GETATTR, self._cobra_name, name)

        if mtype == COBRA_ERROR:
            raise data

        return data

    # For use with ref counted proxies
    def __enter__(self):
        return self

    def __exit__(self, extype, value, tb):
        with self._cobra_getsock() as csock:
            ok = True
            if extype is not None: # Tell the server we broke...
                ok = False
            csock.cobraTransaction(COBRA_GOODBYE, self._cobra_name, ok)

def addSocketBuilder( host, port, builder ):
    '''
    Register a global socket builder which should be used
    when constructing sockets to the given host/port.
    '''
    socket_builders[ (host,port) ] = builder

def getSocketBuilder(host, port):
    '''
    Retrieve the registered socket builder for the given host/port.
    '''
    return socket_builders.get((host,port))

def initSocketBuilder(host,port):
    '''
    Retrieve or initialize a socket builder for the host/port.
    '''
    builder = socket_builders.get((host,port))
    if builder is None:
        builder = SocketBuilder(host,port)
        socket_builders[ (host,port) ] = builder
    return builder

def getCobraDaemon(host="", port=COBRA_PORT, sslca=None, sslcrt=None, sslkey=None, msgpack=True, create=True, wait=False):
    '''
    getCobraDaemon starts a CobraDaemon with the provided arguments and 
    launches it in a separate thread.  The CobraDaemon object is returned
    to the caller.

    If there's already an existing CobraDaemon running for this host/port, 
    that is returned instead.

    This command replaces startCobraServer
    '''
    return daemon_manager.getCobraDaemon(host, port, sslca, sslcrt, sslkey, msgpack)

def startCobraServer(host="", port=COBRA_PORT, sslca=None, sslcrt=None, sslkey=None, msgpack=True):
    '''
    DEPRECATED.  See getCobraDaemon instead.

    startCobraServer starts a CobraDaemon with the provided arguments and 
    launches it in a separate thread.  The CobraDaemon object is returned
    to the caller.

    This API remains for backward compatability only, and duplicates
    getCobraDaemon() functionality
    '''
    return daemon_manager.getCobraDaemon(host, port, sslca, sslcrt, sslkey, msgpack)


class DaemonManager:
    def __init__(self):
        self.daemons = {}

    def get(self, host, port):
        '''
        Simple accessor for the DaemonManager's daemons dict
        Returns None if no daemon is already registered for that host/port
        '''
        return self.daemons.get((host, port))

    def registerCobraDaemon(self, daemon):
        '''
        Registers the given CobraDaemon using the (host, port) combination
        '''
        key = (daemon.host, daemon.port)
        if key in self.daemons:
            raise Exception("Daemon already exists on host %r port %d.  Please deregister that port first." % key)

        self.daemons[key] = daemon

    def deregisterCobraDaemon(self, host, port):
        '''
        Deregisters the CobraDaemon registered for (host, port)
        '''
        if (host, port) in self.daemons:
            self.daemons.pop((host, port))

        else:
            logger.warning("Attempted to deregister a CobraDaemon on a daemon that isn't registered: %r:%d", host, port)


    def getCobraDaemon(self, host="", port=COBRA_PORT, sslca=None, sslcrt=None, sslkey=None, msgpack=True, create=True, wait=False):
        '''
        Returns a registered CobraDaemon for the given parameters.  First checks if
        one is already running for the current host/port and returns that.
        Otherwise, a new CobraDaemon is started and registered
        The CobraDaemon is returned.
        '''
        daemon = self.get(host, port)
        if daemon is None and create:
            daemon = CobraDaemon(host, port, sslca=sslca, sslcrt=sslcrt, sslkey=sslkey, msgpack=msgpack)
            self.registerCobraDaemon(daemon)

            if wait:
                # never returns
                daemon.serve_forever()

        if wait:
            # daemon already exists, join it (still never returns)
            daemon.thr.join()
        else:
            daemon.fireThread()

        return daemon


def shareObject(obj, name=None, doref=False):
    """
    If shareObject is called before startCobraServer
    or startCobraSslServer, it will call startCobraServer
    """
    daemon = getCobraDaemon()
    return daemon.shareObject(obj, name, doref=doref)

def unshareObject(name):
    return daemon.unshareObject(name)

def swapCobraObject(uri, newname):
    '''
    Parse out the object name from a given cobra
    URI and return a newly constructed URI for
    the shared object <newname> on the same server.
    '''
    scheme, host, port, name, urlparams = chopCobraUri( uri )
    paramstr = ''
    if urlparams:
        paramstr = '?' + ('&'.join(['%s=%s' % (k,v) for (k,v) in urlparams.items()]))
    return '%s://%s:%d/%s%s' % (scheme,host,port,newname,paramstr)

def requireMsgpack():
    try:
        import msgpack
    except ImportError:
        raise Exception('Missing "msgpack" python module ( http://visi.kenshoto.com/viki/Msgpack )')

daemon_manager = DaemonManager()




"""
A module full of utils for vectored input tracking and code
flow analysis. (when a scalpel finds something you need to be
able to figure out how to get to it right?)
"""

import vivisect.exc as viv_exc
import vivisect.tools.graphutil as v_graphutil
import vivisect.impemu.monitor as viv_imp_monitor

import visgraph.pathcore as vg_path

from vivisect.const import *

import envi

class InputMonitor(viv_imp_monitor.EmulationMonitor):

    def __init__(self):
        viv_imp_monitor.EmulationMonitor.__init__(self)
        self.res = []

    def apicall(self, emu, op, pc, api, argv):
        magicargs = [(val, emu.getMagic(val)) for val in argv]
        self.res.append( (op.va, magicargs) )

def getEmuAtVa(vw, va, maxhit=None):
    """
    Build and run an emulator to the given virtual address
    from the function entry point.

    (most useful for state analysis.  kinda heavy though...)
    """
    fva = vw.getFunction(va)
    if fva is None:
        return None

    cbva,cbsize,cbfva = vw.getCodeBlock(va)
    fgraph = v_graphutil.buildFunctionGraph(vw, fva)

    # Just take the first one off the iterator...
    for path in v_graphutil.getCodePathsTo(fgraph, cbva):

        emu = vw.getEmulator()
        opcodes = v_graphutil.getOpsFromPath(vw, fgraph, path)
        for op in opcodes:
            if op.va == va:
                break

            emu.executeOpcode(op)

        return emu

def trackImportInputs(vw, iname, maxhit=None):
    """
    Works just like trackFunctionInputs but finds calls to
    imports by name instead...
    """
    mon = InputMonitor()
    for va in vw.getImportCallers(iname):
        emu = getEmuAtVa(vw, va, maxhit=maxhit)
        if emu is None:
            continue

        # Set an emulation monitor and step over the call
        emu.setEmulationMonitor(mon)
        emu.stepi()

    return mon.res

def trackFunctionInputs(vw, fva, maxhit=None):
    """
    Find all the callers to the given function and return a list
    of (callva, [ (argval, magic), ...]) tuples.
    """
    mon = InputMonitor()
    for va in vw.getCallers(fva):

        if not vw.getFunction(va):
            vw.vprint('code at 0x%.8x is not part of a function!' % va)
            continue

        emu = getEmuAtVa(vw, va, maxhit=maxhit)
        # Set an emulation monitor and step over the call
        emu.setEmulationMonitor(mon)
        emu.stepi()

    return mon.res

def trackArgOrigin(vw, fva, argidx):
    """
    Return an input tree (visgraph path tree) of the trackable inputs
    to the specified function.

    Each node in the list will be a leaf node for a path leading
    down toward a call to the target function.  Each node will have
    the following path node properties:

    fva    - The function
    argidx - The index of the argument input with this call
    cva    - The address of the call (to our next) (None on root node)
    argv   - A list of (<val>,<magic>) tuples for the call args (None on root node)
    """

    rootpath = vg_path.newPathNode(fva=fva, cva=None, trackidx=argidx, argidx=None, argv=None)

    todo = [rootpath, ]

    while len(todo):

        path = todo.pop()

        fva = vg_path.getNodeProp(path, 'fva')
        trackidx = vg_path.getNodeProp(path, 'trackidx')

        # Get all of our callers and their arguments to us
        for callva, argv in trackFunctionInputs(vw, fva):

            newfva = vw.getFunction(callva)
            pargs = dict(parent=path, fva=newfva, cva=callva, argidx=trackidx, argv=argv)
            newpath = vg_path.newPathNode(**pargs)

            aval, amagic = argv[trackidx]
            if isinstance(amagic, viv_magic.StackArg) and newfva:
                vg_path.setNodeProp(newpath, 'trackidx', amagic.index)
                todo.append(newpath)

    return vg_path.getLeafNodes(rootpath)

def getCodeFlow(vw, cbva):
    """
    Get a list of the code blocks which are known to flow
    into this one.  This *will* cross function boundaries.
    """
    ret = []
    # Get our actual xrefs
    for fromva, tova, xtype, xdata in vw.getXrefsTo(cbva, REF_CODE):
        xcb = vw.getCodeBlock(fromva)
        if xcb is not None:
            ret.append(xcb)

    # Lets see if the instruction before this was a fallthrough
    ploc = vw.getPrevLocation(cbva)
    if ploc is not None:
        pva, psize, ptype, pinfo = ploc
        # If it's an opcode with fallthrough, count this one too...
        if ptype == LOC_OP and not pinfo & envi.IF_NOFALL:
            pblock = vw.getCodeBlock(pva)
            if pblock is not None:
                ret.append(pblock)

    return ret

def getCodePaths(vw, fromva, tova, trim=True):
    """
    Return a list of paths, where each path is a list
    of code blocks from fromva to tova.

    Usage: getCodePaths(vw, <fromva>, <tova>) -> [ [frblock, ..., toblock], ...]

    NOTE: "trim" causes an optimization which may not reveal *all* the paths,
          but is much faster to run.  It will never return no paths when there
          are some, but may not return all of them... (based on path overlap)
    """

    done = {}
    res = []

    frcb = vw.getCodeBlock(fromva)
    tocb = vw.getCodeBlock(tova)
    if frcb is None:
        raise viv_exc.InvalidLocation(fromva)
    if tocb is None:
        raise viv_exc.InvalidLocation(tova)

    frva = frcb[0] # For compare speed

    root = vg_path.newPathNode(cb=tocb, cbva=tocb[0])
    todo = [root, ]
    done[tova] = tocb

    cbcache = {}

    while len(todo):

        path = todo.pop()
        cbva = vg_path.getNodeProp(path, 'cbva')

        codeblocks = cbcache.get(cbva)
        if codeblocks is None:
            codeblocks = getCodeFlow(vw, cbva)
            cbcache[cbva] = codeblocks

        for cblock in codeblocks:

            bva,bsize,bfva = cblock

            # Don't follow loops...
            if vg_path.isPathLoop(path, 'cbva', bva):
                continue

            # If we have been here before and it's *not* the answer,
            # skip out...
            if trim and done.get(bva) is not None:
                continue

            done[bva] = cblock

            newpath = vg_path.newPathNode(parent=path, cb=cblock, cbva=bva)

            # If this one is a match, we don't need to
            # track past it.  Also, put it in the results list
            # so we don't have to do it later....
            if bva == frva:
                #res.append(newpath)
                fullpath = vg_path.getPathToNode(newpath)

                # We actually do it by inbound references, so reverse the result!
                fullpath.reverse()
                yield [vg_path.getNodeProp(path, 'cb') for path in fullpath]
            else:
                todo.append(newpath)





from PyQt5.QtWidgets import QWidget, QTextEdit, QDialogButtonBox, QMessageBox

from vqt.basics import *

example_src = '''
// A comment (not /* yet... */ )
// "pure" c types only for now...
struct example {
    char x; // This is ok too...
    unsigned int y[20];
};
'''


class UserStructEditor(QWidget):

    def __init__(self, vw, name=None, parent=None):
        '''
        Open a view to edit/create user structure defs.
        If "name" is none, we assume they are creating one.
        '''
        QWidget.__init__(self, parent=parent)
        self._v_vw = vw
        self._v_sname = name
        self._v_changed = True

        ssrc = example_src
        if name:
            ssrc = vw.getUserStructSource(name)
            self._v_changed = False

        self.srcedit = QTextEdit(parent=self)
        self.srcedit.setText(ssrc)
        self.srcedit.textChanged.connect(self._text_changed)

        buttons = QDialogButtonBox(QDialogButtonBox.Save, parent=self)
        buttons.accepted.connect(self._save_event)

        self._set_title()
        self.setLayout(VBox(self.srcedit, buttons))

    def _set_title(self):

        name = self._v_sname
        if name is None:
            name = '(new)'

        status = ''
        if self._v_changed:
            status = '(unsaved!)'

        self.setWindowTitle('Struct Edit: %s %s' % (name, status))

    def _save_event(self):

        ssrc = str(self.srcedit.toPlainText())

        try:
            self._v_sname = self._v_vw.setUserStructSource(ssrc)
            self._v_changed = False
            self._set_title()
        except Exception as e:
            QMessageBox.warning(self, 'Syntax Error', str(e), QMessageBox.Ok)

    def _text_changed(self):
        self._v_changed = True
        self._set_title()



'''
Unified CLI code for things like vivisect and vdb.
'''

import os
import re
import sys
import code
import json
import shlex
import logging
import binascii
import optparse
import traceback
import threading
import collections

import envi.bits as e_bits
import envi.memory as e_mem
import envi.common as e_common
import envi.config as e_config
import envi.memcanvas as e_canvas
import envi.expression as e_expr
import envi.symstore.resolver as e_resolv
import envi.memcanvas.renderers as e_render

from cmd import Cmd
from getopt import getopt


logger = logging.getLogger(__name__)

def splitargs(cmdline):
    cmdline = cmdline.replace('\\\\"', '"').replace('\\"', '')
    patt = re.compile(r'".+?"|\S+')
    for item in cmdline.split('\n'):
        return [s.strip('"') for s in patt.findall(item)]

def formatargs(args):
    ret = []
    subcmds = [args[i:i+5] for i in range(0, len(args), 5)]
    for cmdnames in subcmds:
        fmtstr = '{:15}' * len(cmdnames)
        yield fmtstr.format(*cmdnames)

def columnstr(slist):
    msize = 0
    for s in slist:
        if len(s) > msize:
            msize = len(s)
    return [x.ljust(msize) for x in slist]


class CliExtMeth:
    """
    This is used to work around the difference
    between functions and bound methods for extended
    command modules
    """
    def __init__(self, cli, func):
        self.cli = cli
        self.func = func
        self.__doc__ = func.__doc__

    def __call__(self, line):
        return self.func(self.cli, line)

def isValidScript(scriptpath):
    '''
    Takes in a filepath
    Returns whether the file is valid python (ie. suvives import)
    '''
    if not os.path.isfile(scriptpath):
        return False

    with open(scriptpath, 'rb') as f:
        contents = f.read()

    try:
        cobj = compile(contents, scriptpath, 'exec')
        return True
    except Exception:
        pass

    return False

def getRelScriptsFromPath(scriptpaths):
    '''
    Takes in a list of base paths (eg. ENVI_SCRIPT_PATH list) and recurses the
    directories looking for valid python files (ie. they don't throw errors
    on import).

    Returns a list of scripts usable from the cli in *relative path* format.
    ie.  if my path has "/home/hacker/fooscripts" in it, the script located
    at "/home/hacker/fooscripts/barmazing/bazthis.py" is listed as
    "barmazing/bazthis.py" and the do_script() handler can use that.
    '''
    scripts = []
    for basedir in scriptpaths:
        baselen = len(basedir)

        for dirname,subdirs,subfiles in os.walk(basedir):
            for subfile in subfiles:
                subpath = os.path.join(dirname,subfile)
                if isValidScript(subpath):
                    script = subpath[baselen:]
                    if script.startswith(os.sep):
                        script = script[1:]
                    scripts.append(script)

    return scripts

cfgdefs = {
    'cli':{
        'verbose':False,
        'aliases':{
        }
    }
}

class VOptionParser(optparse.OptionParser):
    '''
    overloads error function that prints to stdout/stderr.

    error is overloaded to raise an exception if an error occurs during the
    parse of arguments.  normally optionparser sends it to stderr.
    '''
    def __init__(self, *args, **kwargs):
        optparse.OptionParser.__init__(self, *args, add_help_option=False, **kwargs)

    def error(self, msg):
        raise Exception(msg)

class EnviCli(Cmd):

    def __init__(self, memobj, config=None, symobj=None):

        self.extcmds = {}
        self.basecmds = []
        self.emptymeth = None
        self.extsubsys = collections.defaultdict(list)
        self.scriptpaths = []
        self.addScriptPathEnvVar('ENVI_SCRIPT_PATH')

        Cmd.__init__(self, stdout=self)

        for name in dir(self):
            if name.startswith('do_'):
                self.basecmds.append(name[3:])

        self.shutdown = threading.Event()

        # If they didn't give us a resolver, make one.
        if symobj is None:
            symobj = e_resolv.SymbolResolver()

        if config is None:
            config = e_config.EnviConfig(defaults=cfgdefs)

        # Force it to be there if its not
        config.getSubConfig('cli')

        self.config = config
        self.memobj = memobj
        self.symobj = symobj
        self.canvas = e_canvas.MemoryCanvas(memobj, syms=symobj)

        self.aliases = {} # For *runtime* aliases only!

    def addCmdAlias(self, alias, cmd, persist=False):
        '''
        Add a command alias and optionally save it.

        Specify persist=True to save the alias.

        Example:
            cli.addRuntimeAlias('woot', 'woot -F -T')
        '''
        if not persist:
            self.aliases[ alias ] = cmd
            return
        self.config.cli.aliases[ alias ] = cmd

    def addScriptPathEnvVar(self, pathenv):
        '''
        Reads a script environment variable in, parses it, and stores the paths
        '''
        scriptdirs = os.getenv( pathenv )
        if scriptdirs is not None:
            for scriptdir in scriptdirs.split(os.pathsep):
                if scriptdir in self.scriptpaths:
                    continue

                self.scriptpaths.append( scriptdir )

    def setCanvas(self, canvas):
        """
        Set a new canvas for the CLI and add all the current renderers
        to the new one.
        """
        for name in self.canvas.getRendererNames():
            canvas.addRenderer(name, self.canvas.getRenderer(name))
        self.canvas = canvas

    def write(self, data):
        # For stdout/stderr
        self.canvas.write(data)

    def get_names(self):
        ret = []
        ret.extend(Cmd.get_names(self))
        ret.extend(self.extcmds.keys())
        return ret

    def getExpressionLocals(self):
        """
        Over-ride this to have things like the eval command
        and the python command use more locals than the sybolic
        defaults.
        """
        return e_expr.MemoryExpressionLocals(self.memobj, symobj=self.symobj)

    def registerCmdExtension(self, func, subsys='extended'):
        self.extcmds["do_%s" % func.__name__] = CliExtMeth(self, func)
        self.extsubsys[ subsys ].append( func.__name__ )

    def vprint(self, msg, addnl=True):
        '''
        Print output to the CLI's output handler.  This allows routines to
        print to the terminal or the GUI depending on which mode we're in.

        Example:
            vprint('hi mom!')
        '''
        if addnl:
            msg = msg + "\n"
        self.canvas.write(msg)

    def __getattr__(self, name):
        func = self.extcmds.get(name, None)
        if func is None:
            raise AttributeError(name)
        return func

    def aliascmd(self, line):
        # Check the "runtime" aliases first
        for alias, cmd in self.aliases.items():
            if line.startswith(alias):
                return line.replace(alias, cmd)

        # Now the "configured" aliases
        for alias, cmd in self.config.cli.aliases.items():
            if line.startswith(alias):
                return line.replace(alias, cmd)

        return line

    def cmdloop(self, intro=None):
        if intro is not None:
            self.vprint(intro)

        while not self.shutdown.is_set():
            try:
                Cmd.cmdloop(self, intro=intro)
            except Exception:
                logger.error(traceback.format_exc())

    def emptyline(self):
        return self.do_help('')

    def setEmptyMethod(self, callback):
        '''
        Set a method to be called back in the event of emptyline().
        NOTE: this method is cleared on every onecmd() call.

        ( the method is called with no args )

        Example:

            def do_foo(self, line):

                x = 10
                def showx():
                    print('X: %d' % x)
                    x += 10

                showx()
                cli.setcrmeth(showx)

        '''
        self.emptymeth = callback

    def onecmd(self, line):

        # check for empty line and emptymeth
        if not line.strip() and self.emptymeth:
            self.emptymeth()
            return

        self.emptymeth = None
        lines = line.split("&&")
        try:
            for line in lines:
                line = self.aliascmd(line)
                Cmd.onecmd(self, line)
        except SystemExit:
            raise
        except Exception as msg:
            if self.config.cli.verbose:
                self.vprint(traceback.format_exc())
            self.vprint("\nERROR: (%s) %s" % (msg.__class__.__name__, msg))
            logger.warning("\nERROR: (%s) %s", msg.__class__.__name__, msg, exc_info=1)

        if self.shutdown.is_set():
            return True

    def do_help(self, line):
        if line:
            return Cmd.do_help(self, line)

        self.basecmds.sort()
        self.vprint('\nbasics:')

        for line in formatargs(self.basecmds):
            self.vprint(line)

        subsys = list(self.extsubsys.keys())
        subsys.sort()

        for sub in subsys:
            self.vprint('\n%s:' % sub)
            cmds = self.extsubsys.get(sub)
            cmds.sort()
            for line in formatargs(cmds):
                self.vprint(line)

        self.vprint('\n')

    def do_clear(self, line):
        '''
        Clears the CLI output. (GUI only)
        '''
        self.canvas.clearCanvas()

    def do_EOF(self, line):
        self.vprint("Use quit")

    def do_quit(self, line):
        """
        Quit

        Usage: quit
        """
        self.shutdown.set()

    def do_config(self, line):
        '''
        Show, edit, or save config options from the command line.

        Usage: config [-s] [config option[=value]]

        no options  display config
        -s          save config to default location AFTER setting any options.
        '''
        parser = VOptionParser()
        parser.add_option('-s', action='store_true', dest='do_save')

        argv = shlex.split(line)
        try:
            options, args = parser.parse_args(argv)
        except Exception as e:
            self.vprint(repr(e))
            return self.do_help('config')

        if len(args) <= 0 and not options.do_save:
            self.vprint(self.config.reprConfigPaths())
            return

        # 1 option per run
        if len(args) > 1:
            return self.do_help('config')

        if len(args) == 1:
            parts = args[0].split('=', 1)
            subnames = parts[0].split('.')

            cfg = self.config
            for subname in subnames[:-1]:
                cfg = cfg.getSubConfig(subname, add=False)
                if cfg is None:
                    self.vprint('No Such Config Section: %s' % subname)
                    return

            optname = subnames[-1]
            optval = cfg.get(optname)
            if optval is None:
                self.vprint('No Such Config Option: %s' % optname)
                return

            if len(parts) == 2:
                # the config entry already has a value, let's use it to decide
                # whether to convert it to an int or leave it as a str.
                if type(cfg[optname]) == int:
                    newval = int(parts[1], 0)
                else:
                    newval = parts[1]

                if (not isinstance(newval, str)) or not isinstance(optval, str):
                    if type(newval) != type(optval):
                        self.vprint('Invalid Type Mismatch: %r - %r' % (newval, optval))
                        return

                optval = newval
                cfg[optname] = newval

            self.vprint('%s.%s=%s' % ('.'.join(subnames), optname, json.dumps(optval)))

        if options.do_save:
            self.config.saveConfigFile()
            self.vprint('saved configuration file to: %s' % self.config.filename)

    def do_alias(self, line):
        """
        Add an alias to the command line interpreter's aliases dictionary
        Usage: alias <alias_word> rest of the alias command

        To delete an alias:
        Usage: alias <alias_word>
        """
        if len(line):
            row = line.split(None, 1)
            self.config.cli.aliases.pop(row[0])
            if len(row) > 1:
                self.config.cli.aliases[ row[0] ] = row[1]

        self.vprint('')
        self.vprint('Runtime Aliases (not saved):')
        aliases = list(self.aliases.keys())
        aliases.sort()
        for alias in aliases:
            self.vprint('%s -> %s' % (alias,self.aliases.get(alias)))
        self.vprint('')

        self.vprint('Configured Aliases:')
        aliases = list(self.config.cli.aliases.keys())
        aliases.sort()
        for alias in aliases:
            self.vprint('%s -> %s' % (alias, self.config.cli.aliases.get(alias)))
        self.vprint('')
        return

    def do_python(self, line):
        """
        Start an interactive python interpreter. The namespace of the
        interpreter is updated with expression nicities.  You may also
        specify a line of python code as an argument to be exec'd without
        beginning an interactive python interpreter on the controlling
        terminal.

        Usage: python [pycode]
        """
        locals = self.getExpressionLocals()
        if len(line) != 0:
            cobj = compile(line, 'cli_input', 'exec')
            exec(cobj, locals)
        else:
            code.interact(local=locals)

    def parseExpression(self, expr):
        return int(e_expr.evaluate(expr, self.getExpressionLocals()))

    def do_binstr(self, line):
        '''
        Display a binary representation of the given value expression
        (padded to optional width in bits)

        Usage: binstr <val_expr> [<bitwidth_expr>]
        '''
        argv = splitargs(line)
        if len(argv) == 0:
            return self.do_help('binstr')
        bitwidth = None
        value = self.parseExpression(argv[0])
        if len(argv) > 1:
            bitwidth = self.parseExpression(argv[1])
        binstr = e_bits.binrepr(value, bitwidth=bitwidth)
        self.canvas.addText("0x%.8x (%d) %s\n" % (value, value, binstr))

    def do_eval(self, line):
        """
        Evaluate an expression on the CLI to show it's value.

        Usage: eval (ecx+edx)/2
        """
        if not line:
            return self.do_help("eval")

        value = self.parseExpression(line)

        self.canvas.addText("%s = " % line)
        if self.memobj.isValidPointer(value):
            self.canvas.addVaText("0x%.8x" % value, value)
            sym = self.symobj.getSymByAddr(value, exact=False)
            if sym is not None:
                self.canvas.addText(" ")
                self.canvas.addVaText("%s + %d" % (repr(sym),value-int(sym)), value)
        else:
            self.canvas.addText("0x%.8x (%d)" % (value, value))

        self.canvas.addText("\n")

    def do_script(self, line):
        '''
        Execute a python file.

        The script file is arbitrary python code which is run with the
        full complement of expression extensions mapped in as locals.

        The script command sources the env var ENVI_SCRIPT_PATH.

        NOTE: additional command line arguments may be passed in and will
              appear as the list "argv" in the script namespace!  (They will
              all be strings)

        Usage: script <scriptfile> [<argv[0]>, ...]

        or     script ?
        '''
        if len(line) == 0:
            return self.do_help('script')

        argv = splitargs(line)
        locals = self.getExpressionLocals()
        locals['argv'] = argv

        if len(argv) and argv[0] == "?":
            scripts = getRelScriptsFromPath(self.scriptpaths)
            scripts.sort()
            self.vprint('Scripts available in script paths:\n\t' + '\n\t'.join(scripts))
            return


        # TODO: unify vdb.extensions.loadExtensions VDB_EXT_PATH with this
        # TODO: where should env var parsing live?
        scriptpath = None
        if os.path.exists(argv[0]):
            scriptpath = argv[0]
        else:
            for scriptdir in self.scriptpaths:
                # allow scripts to import things from the script dir
                if scriptdir not in sys.path:
                    sys.path.append(scriptdir)

                spath = os.path.join(scriptdir, argv[0])
                if os.path.exists(spath):
                    scriptpath = spath

        if scriptpath is None:
            self.vprint('failed to find script')
            return

        with open(scriptpath, 'rb') as f:
            contents = f.read()

        try:
            cobj = compile(contents, scriptpath, 'exec')
            exec(cobj, locals)
        except Exception as e:
            self.vprint( traceback.format_exc() )
            self.vprint('SCRIPT ERROR: %s' % e)

    def do_maps(self, line):
        """
        Display either a list of all the memory maps or the memory map
        details for the given address expression.

        Usage: maps [addr_expression]
        """
        argv = splitargs(line)
        if len(argv):
            expr = " ".join(argv)
            va = self.parseExpression(expr)
            map = self.memobj.getMemoryMap(va)
            if map is None:
                self.vprint("Memory Map Not Found For: 0x%.8x"%va)

            else:
                addr,size,perm,fname = map
                pname = e_mem.reprPerms(perm)
                self.canvas.addText("Memory Map For: ")
                self.canvas.addVaText("0x%.8x" % va, va)
                self.canvas.addText("\n")
                self.canvas.addVaText("0x%.8x" % addr, addr)
                self.canvas.addText("\t%d\t%s\t%s\n" % (size,pname,fname))
        else:
            totsize = 0
            self.vprint("[ address ] [ size ] [ perms ] [ File ]")
            for addr,size,perm,fname in self.memobj.getMemoryMaps():
                pname = e_mem.reprPerms(perm)
                totsize += size
                self.canvas.addVaText("0x%.8x" % addr, addr)
                sizestr = ("%dK" % (size//1024,)).rjust(8)
                self.canvas.addText("%s\t%s\t%s\n" % (sizestr,pname,fname))
            self.vprint("Total Virtual Memory: %.2f MB" % ((float(totsize)/1024)/1024))

    def do_saveout(self, line):
        '''
        saves output to file for any command.  still outputs to whatever
        canvas the command normally outputs to.

        saveout <output file> <cli command>

        Example:
        saveout out.txt search -c MZ
        '''
        argv = shlex.split(line)
        if len(argv) < 2:
            return self.do_help('saveout')

        fname = argv[0]
        command = ' '.join(argv[1:])

        strcanvas = e_canvas.FileBackedMemoryCanvas(fname, self.canvas.mem)
        with e_canvas.TeeCanvas(self, (self.canvas, strcanvas)) as tc:
            self.onecmd(command)

        strcanvas.flush()

    def do_search(self, line):
        '''
        search memory for patterns.

        search [options] <pattern>

        -e  <codec> encode the pattern with a codec (hex, utf-16le, etc)
        -X  pattern is in hex (ie. 41414242 is AABB)
        -E  pattern is an envi memory expression (numeric search)
        -r  pattern is a regular expression
        -R  <baseexpr:sizeexpr> search a range of memory (base + size)
        -c  show context (32 bytes) after each hit
        '''
        parser = VOptionParser()
        parser.add_option('-e', action='store', dest='encode_as')
        parser.add_option('-X', action='store_true', dest='is_hex')
        parser.add_option('-E', action='store_true', dest='is_expr')
        parser.add_option('-r', action='store_true', dest='is_regex')
        parser.add_option('-R', action='store', dest='range_search')
        parser.add_option('-c', action='store_const', dest='num_context_bytes',
                const=32)

        argv = shlex.split(line)
        try:
            options, args = parser.parse_args(argv)
        except Exception as e:
            self.vprint(repr(e))
            return self.do_help('search')

        pattern = (' '.join(args)).encode('utf-8')
        if len(pattern) == 0:
            self.vprint('you must specify a pattern')
            return self.do_help('search')

        if options.is_expr:
            sval = self.parseExpression(pattern)
            endian = self.memobj.getEndian()
            size = self.memobj.getPointerSize()
            pattern = e_bits.buildbytes(sval, size, bigend=endian)

        if options.is_hex:
            pattern = binascii.unhexlify(pattern)

        if options.encode_as is not None:
            if options.encode_as == 'hex':
                pattern = e_common.hexify(pattern)
            else:
                import codecs
                patternutf8 = pattern.decode('utf-8')
                pattern = codecs.encode(patternutf8, encoding=options.encode_as)

        if options.range_search:
            try:
                addrexpr, sizeexpr = options.range_search.split(":")
            except Exception as e:
                self.vprint(repr(e))
                return self.do_help('search')
            addr = self.parseExpression(addrexpr)
            size = self.parseExpression(sizeexpr)

            self.canvas.addText('searching from ')
            self.canvas.addVaText('0x%.8x' % addr, addr)
            self.canvas.addText(' for %d bytes\n' % size)
            res = self.memobj.searchMemoryRange(pattern, addr, size, regex=options.is_regex)
        else:
            self.vprint('searching all memory...')
            res = self.memobj.searchMemory(pattern, regex=options.is_regex)

        if len(res) == 0:
            self.vprint('pattern not found: %s (%s)' % (e_common.hexify(pattern), repr(pattern)))
            return

        brend = e_render.ByteRend()
        self.vprint('matches for: %s (%s)' % (e_common.hexify(pattern), repr(pattern)))
        for va in res:
            mbase,msize,mperm,mfile = self.memobj.getMemoryMap(va)
            pname = e_mem.reprPerms(mperm)
            sname = self.reprPointer(va)

            self.canvas.addVaText('0x%.8x' % va, va)
            self.canvas.addText(': ')
            self.canvas.addText('%s ' % pname)
            self.canvas.addText(sname)

            if options.num_context_bytes is not None:
                self.canvas.addText('\n')
                self.canvas.renderMemory(va, options.num_context_bytes, rend=brend)

            self.canvas.addText('\n')

        self.vprint('done (%d results).' % len(res))

    def reprPointer(self, va):
        """
        Do your best to create a humon readable name for the
        value of this pointer.
        """
        if va == 0:
            return "NULL"

        try:
            mbase, msize, mperm, mfile = self.memobj.getMemoryMap(va)
            if va == mbase:
                ret = mfile
            else:
                ret = mfile + " + 0x%x" % (va - mbase)

            sym = self.symobj.getSymByAddr(va, exact=False)
            if sym is not None:
                ret = "%s + 0x%x" % (repr(sym), va-int(sym))

        except Exception:
            ret = hex(va)

        return ret

    def do_memdump(self, line):
        """
        Dump memory to a file.  If no size is given, the entire memory map that
        contains the given va is dumped to disk.

        Usage: memdump <va_expression> <filename> [size_expression]
        """
        argv = shlex.split(line)
        if len(argv) not in (2, 3):
            return self.do_help('memdump')

        va = self.parseExpression(argv[0])
        fname = argv[1]

        if len(argv) == 2:
            va, size, perm, name = self.memobj.getMemoryMap(va)
        elif len(argv) == 3:
            size = self.parseExpression(argv[2])
        else:
            return self.do_help('memdump')

        mem = self.memobj.readMemory(va, size)
        with open(fname, 'wb') as f:
            f.write(mem)

        self.vprint('wrote %d bytes.' % len(mem))

    def do_memcmp(self, line):
        '''
        Compare memory at the given locations.  Outputs a set of
        differences showing bytes at their given offsets....

        Usage: memcmp <addr_expr1> <addr_expr2> <size_expr>
        '''
        if len(line) == 0:
            return self.do_help('memcmp')

        argv = splitargs(line)
        if len(argv) != 3:
            return self.do_help('memcmp')

        addr1 = self.parseExpression(argv[0])
        addr2 = self.parseExpression(argv[1])
        size  = self.parseExpression(argv[2])

        bytes1 = self.memobj.readMemory(addr1, size)
        bytes2 = self.memobj.readMemory(addr2, size)

        res = e_mem.memdiff(bytes1, bytes2)
        if len(res) == 0:
            self.vprint('No Differences!')
            return

        for offset, offsize in res:
            diff1 = addr1+offset
            diff2 = addr2+offset
            self.canvas.addText('==== %d byte difference at offset %d\n' % (offsize,offset))
            self.canvas.addVaText("0x%.8x" % diff1, diff1)
            self.canvas.addText(":")
            self.canvas.addText(e_common.hexify(bytes1[offset:offset+offsize]))
            self.canvas.addText('\n')
            self.canvas.addVaText("0x%.8x" % diff2, diff2)
            self.canvas.addText(":")
            self.canvas.addText(e_common.hexify(bytes2[offset:offset+offsize]))
            self.canvas.addText('\n')

    def do_mem(self, line):
        """
        Show some memory (with optional formatting and size)

        Usage: mem [-F <format>] <addr expression> [size]
        Usage: <enter> ( show next memory chunk after previous mem cmd )

        NOTE: use -F ? for a list of the formats

        """
        fmtname = "bytes"

        if len(line) == 0:
            return self.do_help("mem")

        argv = splitargs(line)
        try:
            opts,args = getopt(argv, "F:")
        except:
            return self.do_help("mem")

        for opt,optarg in opts:
            if opt == "-F":
                fmtname = optarg
                fnames = self.canvas.getRendererNames()

                if fmtname == "?":
                    self.vprint("Registered renderers:")
                    for name in fnames:
                        self.vprint(name)
                    return

                if fmtname not in fnames:
                    self.vprint("Unknown renderer: %s" % fmtname)
                    return

        if len(args) == 0:
            return self.do_help("mem")

        size = 256
        addr = self.parseExpression(args[0])
        if len(args) == 2:
            size = self.parseExpression(args[1])

        scope = {'addr':addr}
        def showmem():
            self.canvas.setRenderer(fmtname)
            self.canvas.renderMemory(scope['addr'], size)
            scope['addr'] += size

        showmem()
        self.setEmptyMethod(showmem)


class EnviMutableCli(EnviCli):
    """
    Cli extensions which require a mutable memory object
    (emulator/trace) rather than a static one (viv workspace)
    """

    def do_memcpy(self, line):
        '''
        Copy memory from one location to another...

        Usage: memcpy <dest_expr> <src_expr> <size_expr>
        '''
        argv = splitargs(line)
        if len(argv) != 3:
            return self.do_help('memcpy')


        dst = self.parseExpression(argv[0])
        src = self.parseExpression(argv[1])
        siz = self.parseExpression(argv[2])

        mem = self.memobj.readMemory(src, siz)
        self.memobj.writeMemory(dst, mem)

    def do_memprotect(self, line):
        """
        Change the memory permissions of a given page/map.

        Usage: memprotect [options] <addr_expr> <perms>
        -S <size> Specify the size of the region to change (default == whole memory map)
        <perms> = "rwx" string "rw", "rx" "rwx" etc...
        """
        if len(line) == 0:
            return self.do_help("memprotect")

        size = None
        argv = splitargs(line)
        try:
            opts, args = getopt(argv, "S:")
        except Exception as e:
            return self.do_help("memprotect")

        for opt,optarg in opts:
            if opt == "-S":
                size = self.parseExpression(optarg)

        if len(args) != 2:
            return self.do_help("memprotect")


        addr = self.parseExpression(args[0])
        perm = e_mem.parsePerms(args[1])

        if size is None:
            map = self.memobj.getMemoryMap(addr)
            if map is None:
                raise Exception("Unknown memory map for 0x%.8x" % addr)
            size = map[1]

        self.memobj.protectMemory(addr, size, perm)

    def do_writemem(self, args):
        """
        Over-write some memory in the target address space.
        Usage: writemem [options] <addr expression> <string>
        -X    The specified string is in hex (ie 414141 = AAA)
        -U    The specified string needs to be unicode in mem (AAA -> 410041004100)
        """
        dohex = False
        douni = False

        try:
            argv = splitargs(args)
            opts,args = getopt(argv, "XU")
        except:
            return self.do_help("writemem")

        if len(args) != 2:
            return self.do_help("writemem")

        for opt,optarg in opts:
            if opt == "-X":
                dohex = True
            elif opt == "-U":
                douni = True

        exprstr, memstr = args
        if dohex:
            memstr = binascii.unhexlify(memstr)
        if douni:
            memstr = ("\x00".join(memstr)) + "\x00"

        addr = self.parseExpression(exprstr)
        self.memobj.writeMemory(addr, memstr)

